// =========================================================================//
//                                                                          //
//  TCaret -  Windows (C)  .. 1998                           //
//                                                                          //
// =========================================================================//

unit Caret;

interface

uses Classes, Windows;

type
  TCaret = class(TObject)
  private
    FHideCount: Longint;
    FHandle: HWND;
    FWidth: Longint;
    FHeight: Longint;
    function  GetVisible: Boolean;
    procedure SetHeight(AHeight: Longint);
    procedure SetWidth(AWidth: Longint);
    procedure InitCaret(AWidth, AHeight, AHideCount: Longint);
  public
    constructor Create(AHandle: HWND; AWidth, AHeight: Longint);
    destructor  Destroy; override;
    procedure   Hide;
    procedure   Show;
    procedure   SetPos(X, Y: Longint);
    procedure   GetPos(var X, Y: Longint);

    property Visible: Boolean read GetVisible;
    property Height: Longint read FHeight write SetHeight;
    property Width:  Longint read FWidth  write SetWidth;
  end;


implementation

// ---------------------------------------------------------------------------
function TCaret.GetVisible: Boolean;
begin
  Result := False;
  if Self = nil then Exit;

  Result := (FHideCount = 0);
end;

// ---------------------------------------------------------------------------
procedure TCaret.InitCaret(AWidth, AHeight, AHideCount: Longint);
begin
  CreateCaret(FHandle, 0, AWidth, AHeight);
  FWidth  := AWidth;
  FHeight := AHeight;
  FHideCount := AHideCount;
end;

// ---------------------------------------------------------------------------
constructor TCaret.Create(AHandle: HWND;  AWidth, AHeight: Longint);
begin
  inherited Create;

  FHandle := AHandle;
  InitCaret(AWidth, AHeight, 1);
end;

// ---------------------------------------------------------------------------
destructor TCaret.Destroy;
begin
  DestroyCaret;

  inherited Destroy;
end;

// ---------------------------------------------------------------------------
procedure TCaret.Hide;
begin
  if Self = nil then Exit;

  HideCaret(FHandle);
  Inc(FHideCount);
end;

// ---------------------------------------------------------------------------
procedure TCaret.Show;
begin
  if Self = nil then Exit;

  ShowCaret(FHandle);
  if (FHideCount > 0) then Dec(FHideCount);
end;

// ---------------------------------------------------------------------------
procedure TCaret.SetPos(X, Y: Longint);
begin
  if Self = nil then Exit;

  SetCaretPos(X, Y);
end;

// ---------------------------------------------------------------------------
procedure TCaret.GetPos(var X, Y: Longint);
var
  P: TPoint;
begin
  if Self = nil then Exit;

  GetCaretPos(P);
  X := P.X;
  Y := P.Y;
end;

// ---------------------------------------------------------------------------
procedure TCaret.SetHeight(AHeight: Longint);
var
  OldVisible: Boolean;
begin
  if Self = nil then Exit;

  OldVisible := Visible;
  DestroyCaret;
  InitCaret(FWidth, AHeight, 1);

  if OldVisible then Show;
end;

// ---------------------------------------------------------------------------
procedure TCaret.SetWidth(AWidth: Longint);
var
  OldVisible: Boolean;
begin
  if Self = nil then Exit;

  OldVisible := Visible;
  DestroyCaret;
  InitCaret(AWidth, FHeight, 1);

  if OldVisible then Show;
end;

end.
