//==========================================================================//
//                                                                          //
//  TColorEdit  1.10 (C)  .. 1998                             //
//                                                                          //
//==========================================================================//

unit ColorEdt;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, SymColor, Caret;

const
  CEM_AFTERENTER = WM_USER + 100;
  CEM_AFTEREXIT  = WM_USER + 101;

type
  //     
  //    , 
  //        ,
  //         (. SetUndoBuffer)
  TUndoRecord = record
    Valid:       Boolean; //    ?
    LastCommand: Longint; //   (  b )
    Text:        String;  //    
    CaretPos:    Longint; // - // -
    TextBegPos:  Longint; // - // -
    TextEndPos:  Longint; // - // -
    SelBegPos:   Longint; // - // -
    SelEndPos:   Longint; // - // -
  end;

  //     
  //            :
  // - PaintNeeded
  TModifyState = record
    Text:        String;
    CaretPos:    Longint;
    TextBegPos:  Longint;
    TextEndPos:  Longint;
    SelBegPos:   Longint;
    SelEndPos:   Longint;
  end;

  //       
  TEditColorEvent = procedure (Sender: TObject; Ch: Char; Index: Longint; Selected: Boolean;
                           var TextColor, BkColor: TColor) of object;

  //    
  TAcceptCharEvent = procedure (Sender: TObject; Ch: Char; var Accept: Boolean) of object;

  // esMouseSelection    -  ()  
  // esDisplaySelection  -   
  // esRetainFocus       -   ( .   . )
  TEditStates = (esMouseSelection, esDisplaySelection, esRetainFocus);
  TEditState = set of TEditStates;

  TColorEdit = class(TCustomControl)
  private
    //  
    FEditState: TEditState;            //   
    FPaintLock: Integer;               //  
    FCaret: TCaret;                    //  
    FCaretPos: Longint;                //   1 - Length(FText);
    FCaretX, FCaretY: Longint;         //   ()
    FTextBegPos, FTextEndPos: Longint; //      
    FShiftState: TShiftState;          //    (Alt, Ctrl, Shift)
    FSelBegPos, FSelEndPos: Longint;   //      
    FUndoBuffer: TUndoRecord;          //    
    FModifyState: TModifyState;

    //    (property)
    FAutoSelect: Boolean;              //     
    FAutoSize: Boolean;                //     
    FBorderStyle: TBorderStyle;        //  
    FCharCase: TEditCharCase;          // // 
    FHideSelection: Boolean;           //      ?
    FMaxLength: Integer;               //   
    FReadOnly: Boolean;                //   
    FText:  String;                    // 
    FTextColor   : TSymbolColor;       //  
    FSelectColor : TSymbolColor;       //   
    FDisableColor: TSymbolColor;       //    Enabled := False
    FModified: Boolean;                //   ""

    //   
    FOnGetSymbolColor: TEditColorEvent;    //  .    
    FOnAcceptChar: TAcceptCharEvent;   //    
    FOnChange   : TNotifyEvent;        //   
    FAfterEnter : TNotifyEvent;        //    
    FAfterExit  : TNotifyEvent;        //    

    //       ,
    //  TControl,       ,
    //    - TextColor  Font
    FOldOnFontChange: TNotifyEvent;

    //  -
    function  GetModified: Boolean;
    procedure SetModified(Value: Boolean);
    procedure SetVersion(Value: String);
    procedure SetTextColor(Value: TSymbolColor);
    procedure SetSelectColor(Value: TSymbolColor);
    procedure SetDisableColor(Value: TSymbolColor);
    procedure SetText(Value: String);
    procedure SetAutoSize(Value: Boolean);
    procedure SetBorderStyle(Value: TBorderStyle);
    procedure SetCharCase(Value: TEditCharCase);
    procedure SetMaxLength(Value: Integer);
    function  GetSelText: String;
    procedure ReplaceSelection(NewSel: String);
    function  GetSelStart: Longint;
    procedure SetSelStart(Value: Longint);
    function  GetSelLength: Longint;
    procedure SetSelLength(Value: Longint);
    procedure SetHideSelection(Value: Boolean);

    function  TextChanged: Boolean;
    procedure RepaintWindow;
    function  PaintNeeded: Boolean;
    procedure ClearModifyState;
    procedure SetSelection(BegPos, EndPos: Longint);
    //      
    function  RightBound: Longint;
    // / 
    procedure PaintLock;
    procedure PaintUnlock;
    // HW:  ?
    function  IsSibling(HW : hWnd): Boolean;
    procedure FontChanged(Sender: TObject);
    procedure TextColorChanged(Sender: TObject);
    procedure SelectColorChanged(SEnder: TObject);
    procedure DisableColorChanged(SEnder: TObject);
    //        
    procedure CalcCaretPosOnMouseEvent(X, Y: Longint);
    //     
    procedure PositionCaret;
    //   > 0 - , < 0 - 
    procedure ScrollText(NumOfChar: Longint);
    //    .  .
    procedure CalcTextBegPos;
    //    .  .
    procedure CalcTextEndPos;
    //
    //   (   ,    )
    //
    procedure CmdGoLeft;
    procedure CmdGoRight;
    procedure CmdGoHome;
    procedure CmdGoEnd;
    //
    //  
    //
    procedure CmdChar(CharCode: Word);
    //   Dir < 0 -   ,  Dir > 0 -   
    procedure CmdDeleteChar(Dir: Longint);
    procedure CmdDeleteSelected;
    procedure CmdSelectWord;
    procedure CmdCopy;
    procedure CmdPaste;
    procedure CmdCut;
    //    
    procedure CmdUndo;

    //         
    procedure CheckSelectMode;
    //    
    procedure ClearSelectMode;
    function  IsSelected: Boolean;
    function  IsAllSelected: Boolean;
    function  InSelectMode: Boolean;
    //  .  .    
    //     
    procedure PerformSelection(OldCaretPos: Longint);
    //        
    procedure   AdjustHeight;
    //  ,         csFixedHeight
    procedure UpdateHeight;
    //   .  ,    
    //  ,  
    procedure AdjustText;
    //   
    procedure SetUndoBuffer(Command: Longint);
    //   
    procedure ClearUndoBuffer;
    //   .     
    // ,      (. SetUndoBuffer)
    procedure SetUndoCommand(Command: Longint);
  protected
    //  
    FVersion: String;

    procedure Paint; override;
    procedure Loaded; override;
    procedure CreateParams(var Params : TCreateParams); override;
    procedure SetName(const Value: TComponentName); override;
    //  Windows
    procedure WMEraseBkGnd(var Msg : TWMEraseBkGnd); message WM_ERASEBKGND;
    procedure WMSetFont(var Msg: TWMSetFont); message WM_SETFONT;
    procedure WMSetFocus(var Msg : TWMSetFocus); message WM_SETFOCUS;
    procedure WMKillFocus(var Msg : TWMKillFocus); message WM_KILLFOCUS;
    procedure WMLButtonDown(var Msg : TWMLButtonDown); message WM_LBUTTONDOWN;
    procedure WMLButtonUp(var Msg : TWMLButtonUp); message WM_LBUTTONUP;
    procedure WMMouseMove(var Msg : TWMMouseMove); message WM_MOUSEMOVE;
    procedure WMKeyDown(var Msg : TWMKeyDown); message WM_KEYDOWN;
    procedure WMGetDlgCode(var Msg : TWMGetDlgCode); message WM_GETDLGCODE;
    procedure WMChar(var Msg : TWMChar); message WM_CHAR;
    procedure WMLButtonDblClk(var Msg : TWMLButtonDblClk); message WM_LBUTTONDBLCLK;
    procedure WMCopy(var Msg : TWMCopy); message WM_COPY;
    procedure WMCut(var Msg : TWMCut); message WM_CUT;
    procedure WMPaste(var Msg : TWMPaste); message WM_PASTE;
    procedure WMClear(var Msg : TWMClear); message WM_CLEAR;
    procedure WMSysKeyDown(var Msg : TWMSysKeyDown); message WM_SYSKEYDOWN;
    procedure WMRButtonUp(var Msg : TWMRButtonDown); message WM_RBUTTONUP;
    //  Delphi
    procedure CMCtl3DChanged(var Msg : TMessage); message CM_CTL3DCHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    //   
    procedure CEMAfterEnter(var Msg : TMessage); message CEM_AFTERENTER;
    procedure CEMAfterExit(var Msg : TMessage);  message CEM_AFTEREXIT;

    //    
    procedure UndoClick(Sender : TObject);      dynamic;
    procedure CutClick(Sender : TObject);       dynamic;
    procedure CopyClick(Sender : TObject);      dynamic;
    procedure PasteClick(Sender : TObject);     dynamic;
    procedure DeleteClick(Sender : TObject);    dynamic;
    procedure SelectAllClick(Sender : TObject); dynamic;

    //     ,  
    //      !
    procedure Change; virtual;
    //    1.10
    property Align;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    //   Clipboard
    procedure   CopyToClipboard;
    procedure   CutToClipboard;
    procedure   PasteFromClipboard;
    //   
    procedure   ClearSelection;
    //   
    procedure   SelectAll;
    //  
    procedure   Clear;

    //    1.10
    procedure ClearUndo;
    procedure Undo;

    //     (. TEdit)
    property SelText: String
             read GetSelText      write ReplaceSelection;
    property SelStart: Longint
             read GetSelStart     write SetSelStart;
    property SelLength: Longint
             read GetSelLength    write SetSelLength;
    //  ? (. TEdit)
    property Modified: Boolean
             read GetModified     write SetModified;
  published
    // 
    property BorderStyle: TBorderStyle
             read FBorderStyle    write SetBorderStyle default bsSingle;
    property Text: String
             read FText           write SetText;
    property HideSelection: Boolean
             read FHideSelection  write SetHideSelection default True;
    property AutoSelect: Boolean
             read FAutoSelect     write FAutoSelect    default True;
    property AutoSize: Boolean
             read FAutoSize       write SetAutoSize    default True;
    property CharCase: TEditCharCase
             read FCharCase       write SetCharCase    default ecNormal;
    property MaxLength: Integer
             read FMaxLength      write SetMaxLength   default 0;
    property ReadOnly: Boolean
             read FReadOnly       write FReadOnly default False;
    property TextColor: TSymbolColor
             read FTextColor      write SetTextColor;
    property SelectColor: TSymbolColor
             read FSelectColor    write SetSelectColor;
    property DisableColor: TSymbolColor
             read FDisableColor   write SetDisableColor;
    property Version: String
             read FVersion        write SetVersion stored False;

    // -   TCustomControl
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;

    //    1.10
    property Anchors;
    property Constraints;
    property DragKind;

    // 
    property OnGetSymbolColor: TEditColorEvent
             read FOnGetSymbolColor     write FOnGetSymbolColor;
    property OnAcceptChar: TAcceptCharEvent
             read FOnAcceptChar   write FOnAcceptChar;
    property OnChange: TNotifyEvent
             read FOnChange       write FOnChange;
    property AfterEnter: TNotifyEvent
             read FAfterEnter     write FAfterEnter;
    property AfterExit: TNotifyEvent
             read FAfterExit      write FAfterExit;

    // ,   TCustomControl;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;

    //    1.10
    property OnEndDock;
    property OnStartDock;

  end;  { TColorEdit }

implementation

uses Menus, Clipbrd;

const
  BorderStyles: array[TBorderStyle] of Longint = (0, WS_BORDER);
  EMPTY_CHAR         = [#32, #9];
  LEFT_MARGIN        = 1;
  RIGHT_MARGIN       = 1;
  TOP_MARGIN         = 1;
  TEXT_SCROLL_CHUNK  = 3;
  CE_DEF_HEIGHT      = 22;
  CE_DEF_WIDTH       = 121;
  CE_DEF_TABSTOP     = True;
  CE_DEF_PARENTCOLOR = False;

  // .    CMD_CHAR      
  CMD_NULL   =  0;
  CMD_COPY   =  3;
  CMD_PASTE  = 22;
  CMD_CUT    = 24;
  CMD_UNDO   = 26;
  CMD_CHAR   = 1001;

  //   
  MI_UNDO_TEXT        = '&';
  MI_CUT_TEXT         = '&';
  MI_COPY_TEXT        = '&';
  MI_PASTE_TEXT       = '&';
  MI_DELETE_TEXT      = '&';
  MI_SELECT_ALL_TEXT  = ' &';

  //  
  COLOR_EDIT_VERSION = '1.10';


function Min (A, B: Longint): Longint;
begin
  Result := A;
  if B < A then Result := B;
end;

function Max (A, B: Longint): Longint;
begin
  Result := A;
  if B > A then Result := B;
end;

{TColorEdit}

//----------------------------------------------------------------------------
//    1.10
//----------------------------------------------------------------------------
procedure TColorEdit.ClearUndo;
begin
  ClearUndoBuffer;
end;

procedure TColorEdit.Undo;
begin
  CmdUndo;
end;

// ---------------------------------------------------------------------------
//  Loaded
//        OnChange,   
//     
// ---------------------------------------------------------------------------

procedure TColorEdit.Loaded;
begin
  inherited Loaded;
  if Assigned(FOnChange) then FOnchange(Self);
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.SetHideSelection(Value: Boolean);
begin
  if Value <> FHideSelection then begin
    FHideSelection := Value;
    Repaint;
  end;
end;

// ---------------------------------------------------------------------------
// GetModified
//     : FUndoBuffer.Valid
//                                "" : FModified
// ---------------------------------------------------------------------------
function  TColorEdit.GetModified: Boolean;
begin
  Result := FModified or FUndoBuffer.Valid;
end;

// ---------------------------------------------------------------------------
// SetModified
//   ""    Change
// ---------------------------------------------------------------------------
procedure TColorEdit.SetModified(Value: Boolean);
begin
  if Value <> FModified then begin
    FModified := Value;
    if not Value then ClearUndoBuffer;
    Change;
  end;
end;

// ---------------------------------------------------------------------------
function  TColorEdit.GetSelLength: Longint;
begin;
  Result := FSelEndPos - FSelBegPos + 1;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.SetSelLength(Value: Longint);
begin;
  if Value <> GetSelLength then begin
    FSelEndPos := FSelBegPos + Value - 1;
    if FSelEndPos > Length(FText) then
      FSelEndPos := Length(FText);
    FCaretPos := FSelEndPos + 1;
    PositionCaret;
    Repaint;
  end;
end;

// ---------------------------------------------------------------------------
function  TColorEdit.GetSelStart: Longint;
begin
  if IsSelected then Result := FSelBegPos // 1.10
  else               Result := FCaretPos;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.SetSelStart(Value: Longint);
begin
  if Value <> FSelBegPos then begin
    SetSelection(0, 0);
    FSelBegPos := Value;
    FSelEndPos := FSelBegPos;
    FCaretPos := FSelBegPos;
    PositionCaret;
    Repaint;
  end;
end;


// ---------------------------------------------------------------------------
function  TColorEdit.TextChanged: Boolean;
begin
  Result :=   (FModifyState.Text <> FText);
end;

// ---------------------------------------------------------------------------
//  RepaintWindow
//       Repaint,  
//        ,  
//      Change.
// ---------------------------------------------------------------------------
procedure TColorEdit.RepaintWindow;
begin
  if PaintNeeded then begin
    if TextChanged then Change;
    Repaint;
  end;
end;

// ---------------------------------------------------------------------------
function TColorEdit.PaintNeeded: Boolean;
begin
  with FModifyState do
    Result := (FModifyState.Text <> FText) or (CaretPos   <> FCaretPos) or
              (TextBegPos <> FTextBegPos) or (TextEndPos <> FTextEndPos) or
              (SelBegPos  <> FSelBegPos)  or (SelEndPos  <> FSelEndPos);

end;

// ---------------------------------------------------------------------------
procedure TColorEdit.ClearModifyState;
begin
  with FModifyState do begin
    FModifyState.Text := FText;
    CaretPos         := FCaretPos;
    TextBegPos       := FTextBegPos;
    TextEndPos       := FTextEndPos;
    SelBegPos        := FSelBegPos;
    SelEndPos        := FSelEndPos;
  end;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.SetVersion(Value: String);
begin
  ; //   !!!
end;

// ---------------------------------------------------------------------------
//  ReplaceSelection
//     
// ---------------------------------------------------------------------------
procedure   TColorEdit.ReplaceSelection (NewSel: String);
begin
  PaintLock; //  CmdDeleteSelected     !
  try
    CmdDeleteSelected;
    Insert(NewSel,  FText, FCaretPos);
    SetSelection(FCaretPos, FCaretPos + Length(NewSel) - 1);
    Inc(FCaretPos, Length(NewSel));
    if FTextEndPos < FSelEndPos then begin
      FTextEndPos := FSelEndPos;
      CalcTextBegPos;
      ScrollText(-TEXT_SCROLL_CHUNK);
    end;
    PositionCaret;
  finally
    PaintUnlock;
  end;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.SetTextColor(Value: TSymbolColor);
begin
  FTextColor.Assign(Value);
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.SetSelectColor(Value: TSymbolColor);
begin
  FSelectColor.Assign(Value);
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.SetDisableColor(Value: TSymbolColor);
begin
  FDisableColor.Assign(Value);
end;

// ---------------------------------------------------------------------------
//  SetName
//       TControl  
//    Text = Name    
// ---------------------------------------------------------------------------
procedure TColorEdit.SetName(const Value: TComponentName);
var
  ChangeText: Boolean;
begin
  ChangeText := (csSetCaption in ControlStyle) and (Name = Text) and
    ((Owner = nil) or not (Owner is TControl) or
    not (csLoading in TControl(Owner).ComponentState));
  inherited SetName(Value);
  if ChangeText then Text := Value;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.FontChanged(Sender: TObject);
begin
  if Assigned(FOldOnFontChange) then
    FOldOnFontChange(Sender);
  if Font.Color <> FTextColor.TextColor then
    FTextColor.TextColor := Font.Color;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.TextColorChanged(Sender: TObject);
begin
 if Font.Color <> FTextColor.TextColor then
    Font.Color := FTextColor.TextColor
  else
    Repaint;  
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.SelectColorChanged(SEnder: TObject);
begin
  Repaint;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.DisableColorChanged(SEnder: TObject);
begin
  Repaint;
end;

// ---------------------------------------------------------------------------
procedure   TColorEdit.ClearSelection;
begin
  SetSelection(0, 0);
  RepaintWindow;
end;

// ---------------------------------------------------------------------------
// IsSibling
//  HW    ?     WM_KILLFOCUS
//       " " esRetainFocus
//  esRetainFocus ,       
//  .   ,     
//      .
// ---------------------------------------------------------------------------
function TColorEdit.IsSibling(HW : hWnd): Boolean;
var
  C : TWinControl;
  H : hWnd;
begin
  Result := False;
  if HW = 0 then Exit;

  C := FindControl(HW);

  if not Assigned(C) then begin
    H := GetParent(HW);
    if H > 0 then
      C := FindControl(H);
  end;

  if Assigned(C) then
    if GetParentForm(C) = GetParentForm(Self) then
      Result := True;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.Change;
begin
  inherited Changed;

  if Assigned(FOnChange) then FOnChange(Self);
end;

// ---------------------------------------------------------------------------
//  PaintLock -  
//  ,       
// ---------------------------------------------------------------------------
procedure TColorEdit.PaintLock;
begin
  Inc(FPaintLock);
end;

// ---------------------------------------------------------------------------
//  PainUnlock -     
// ---------------------------------------------------------------------------
procedure TColorEdit.PaintUnlock;
begin
  if FPaintLock > 0 then begin
    Dec(FPaintLock);
    if FPaintLock = 0 then
      RepaintWindow;
  end;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.SetMaxLength(Value: Integer);
begin
  if (FMaxLength <> Value) and (FMaxLength >= 0) then
  begin
    FMaxLength := Value;
    ClearUndoBuffer;
    if (FMaxLength > 0) and (FMaxLength < Length(FText)) then begin
       FText := Copy(FText, 1, FMaxLength);
       FTextEndPos := FMaxLength;;
       FCaretPos := FTextEndPos + 1;
       CalcTextBegPos;
       SetSelection(0, 0);
       Invalidate;
     end;
  end;
end;

// ---------------------------------------------------------------------------
//  AdjustText -  
//      ,     
// ---------------------------------------------------------------------------
procedure TColorEdit.AdjustText;
var
  OldTextEndPos, OldTextBegPos: Longint;
begin
  //  
  // CalcTextEndPos   FTextEndPos,   !
  OldTextEndPos := FTextEndPos;
  CalcTextEndPos;
  PositionCaret;
  //      ,   
  OldTextBegPos := FTextBegPos;
  if FCaretX > RightBound then begin
    FTextEndPos := OldTextEndPos;
    CalcTextBegPos;
    PositionCaret;
  end;
  if FCaretX > RightBound then //  ,   !
   FTextBegPos := OldTextBegPos;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.SetCharCase(Value: TEditCharCase);
begin
  if FCharCase <> Value then
  begin
    FCharCase := Value;
    case FCharCase of
      ecUpperCase: FText := AnsiUpperCase(FText);
      ecLowerCase: FText := AnsiLowerCase(FText);
    end;
    AdjustText;
    Repaint;
  end;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.SetAutoSize;
begin
 if FAutoSize <> Value then begin
    FAutoSize := Value;
    UpdateHeight;
 end;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.SetBorderStyle(Value: TBorderStyle);
begin
  if FBorderStyle <> Value then begin
    FBorderStyle := Value;
    UpdateHeight;
    RecreateWnd;
  end;
end;

// ---------------------------------------------------------------------------
//  SetUndoCommand -    
//        
//   Undo -  (. SetUndoBuffer)
// ---------------------------------------------------------------------------
procedure TColorEdit.SetUndoCommand(Command: Longint);
begin
  FUndoBuffer.LastCommand := Command;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.ClearUndoBuffer;
begin
  FUndoBuffer.Valid := False;
  FUndoBuffer.LastCommand := CMD_NULL;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.SetUndoBuffer(Command: Longint);
begin
  if (FUndoBuffer.Text <> FText) and
     ((Command = CMD_UNDO) or (FUndoBuffer.LastCommand <> Command)) then begin
    FUndoBuffer.Text := FText;
    with FUndoBuffer do begin
      TextBegPos  := FTextBegPos;
      TextEndPos  := FTextEndPos;
      CaretPos    := FCaretPos;
      SelBegPos   := FSelBegPos;
      SelEndPos   := FSelEndPos;
      LastCommand := Command;
      Valid       := True;
    end;
  end;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.UpdateHeight;
begin
  if FAutoSize and (BorderStyle = bsSingle) then begin
    ControlStyle := ControlStyle + [csFixedHeight];
    AdjustHeight;
  end else
    ControlStyle := ControlStyle - [csFixedHeight];
end;

// ---------------------------------------------------------------------------
//  AdjustHeight -        
//    TCustomEdit
// ---------------------------------------------------------------------------
procedure TColorEdit.AdjustHeight;
var
  DC: HDC;
  SaveFont: HFont;
  I: Integer;
  SysMetrics, Metrics: TTextMetric;
begin
  DC := GetDC(0);
  GetTextMetrics(DC, SysMetrics);
  SaveFont := SelectObject(DC, Font.Handle);
  GetTextMetrics(DC, Metrics);
  SelectObject(DC, SaveFont);
  ReleaseDC(0, DC);
  if NewStyleControls then
  begin
    if Ctl3D then I := 8 else I := 6;
    I := GetSystemMetrics(SM_CYBORDER) * I;
  end else
  begin
    I := SysMetrics.tmHeight;
    if I > Metrics.tmHeight then I := Metrics.tmHeight;
    I := I div 4 + GetSystemMetrics(SM_CYBORDER) * 4;
  end;
  Height := Metrics.tmHeight + I;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.ClearSelectMode;
begin
  Exclude(FShiftState, ssShift);
  Exclude(FEditState, esMouseSelection);
end;

// ---------------------------------------------------------------------------
function TColorEdit.IsAllSelected: Boolean;
begin
  Result := (FSelBegPos = 1) and (FSelEndPos = Length(FText));
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.SelectAll;
begin
  SetSelection(1, Length(FText));
  FTextEndPos := FSelEndPos;
  CalcTextBegPos;
  FCaretPos := FSelEndPos + 1;
  RepaintWindow;
  PositionCaret;
end;

// ---------------------------------------------------------------------------
//  PerformSelection - ""   
//  ,     
// ---------------------------------------------------------------------------
procedure TColorEdit.PerformSelection(OldCaretPos: Longint);
var
  MinSelPos, MaxSelPos: Longint;
begin
  if not InSelectMode then Exit;

  if IsSelected then begin
    MinSelPos := Min(FSelBegPos, FSelEndPos);
    MaxSelPos := Max(FSelBegPos, FSelEndPos);
    if FCaretPos <= MinSelPos then begin
      if FSelEndPos >= FSelBegPos then
        if (FCaretPos = MinSelPos) and (FCaretPos < OldCaretPos) then
          SetSelection(0, 0)
        else if OldCaretPos > MinSelPos  then
          Dec(FSelBegPos);
      if IsSelected then FSelEndPos := FCaretPos;
    end else if FCaretPos > MaxSelPos then begin
      if FSelEndPos <= FSelBegPos then
        if (FCaretPos = MaxSelPos + 1) and (FCaretPos > OldCaretPos) then
          SetSelection(0, 0)
        else if OldCaretPos <= MaxSelPos then
          Inc(FSelBegPos);
      if IsSelected then FSelEndPos := FCaretPos - 1;
    end else begin
      if FSelEndPos > FSelBegPos then
        FSelEndPos := FCaretPos - 1
      else if FSelEndPos < FSelBegPos then
        FSelEndPos := FCaretPos;
    end;
  end else begin
    if FCaretPos > OldCaretPos then begin
      FSelBegPos := OldCaretPos;
      FSelEndPos := FCaretPos - 1;
    end else if FCaretPos < OldCaretPos then begin
      FSelBegPos := OldCaretPos - 1;
      FSelEndPos := FCaretPos;
    end;
  end;
end;

// ---------------------------------------------------------------------------
function  TColorEdit.InSelectMode: Boolean;
begin
  Result := (ssShift in FShiftState) or (esMouseSelection in FEditState);
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.SetSelection(BegPos, EndPos: Longint);
begin
  FSelBegPos := BegPos;
  FSelEndPos := EndPos;
end;

// ---------------------------------------------------------------------------
function  TColorEdit.IsSelected: Boolean;
var
  Len: Longint;
begin
 Len := Length(FText);
 Result := (FSelBegPos > 0) and (FSelEndPos > 0) and
           (FSelBegPos <= Len) and (FSelEndPos <= Len);
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.CheckSelectMode;
begin
  if not InSelectMode then
     SetSelection(0, 0);
end;

// ---------------------------------------------------------------------------
function TColorEdit.RightBound: Longint;
begin
  Result := ClientRect.Right - ClientRect.Left - RIGHT_MARGIN;
end;

// ---------------------------------------------------------------------------
//  ScrollText -   /
// ---------------------------------------------------------------------------
procedure TColorEdit.ScrollText(NumOfChar: Longint);
begin
  if NumOfChar > 0 then begin
    Dec(FTextBegPos, NumOfChar);
    if FTextBegPos < 1 then FTextBegPos := 1;
  end else begin
    Inc(FTextBegPos, -NumOfChar);
    if FTextBegPos > FTextEndPos then
      FTextBegPos := FTextEndPos;
  end;
end;

// ---------------------------------------------------------------------------
//  CalcTextBegPos -    .  .
// ---------------------------------------------------------------------------
procedure TColorEdit.CalcTextBegPos;
var
  I, X: Longint;
begin
  X := LEFT_MARGIN;
  FTextBegPos := 1;
  for I := FTextEndPos downto 1 do begin
    X := X + Canvas.TextWidth(FText[I]);
    if X > RightBound then begin
      if I < Length(FText) then FTextBegPos := I + 1
      else                      FTextBegPos := Length(FText);
      break;
    end;
  end;
end;

// ---------------------------------------------------------------------------
//  CalcTextEndPos -    .  .
//    - Paint      
// ---------------------------------------------------------------------------
procedure TColorEdit.CalcTextEndPos;
var
  I, X: Longint;
begin
  X := LEFT_MARGIN;
  FTextEndPos := Length(FText);
  for I := FTextBegPos to Length(FText) do begin
    X := X + Canvas.TextWidth(FText[I]);
    if X > RightBound then begin
      if I > 1 then FTextEndPos := I - 1
      else          FTextEndPos := 1;
      break;
    end;
  end;
end;

// ---------------------------------------------------------------------------
//  PositionCaret -        () 
//     ().
// ---------------------------------------------------------------------------
procedure TColorEdit.PositionCaret;
var
  I: Longint;
begin
  FCaretX := LEFT_MARGIN;
  for I := FTextBegPos to FCaretPos-1 do
    FCaretX := FCaretX + Canvas.TextWidth(FText[I]);
  FCaret.SetPos(FCaretX, FCaretY);
end;

// ---------------------------------------------------------------------------
//  Create - 
// ---------------------------------------------------------------------------
constructor TColorEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  FVersion := COLOR_EDIT_VERSION;
  FEditState := [esDisplaySelection];
  FReadOnly := False;
  FPaintLock := 0;
  FCaret := nil;
  FMaxLength := 0;
  FCharCase := ecNormal;
  FBorderStyle := bsSingle;
  Canvas.Brush.Color := Color;
  FText := '';
  FCaretX := LEFT_MARGIN;
  FCaretY := TOP_MARGIN;
  FTextBegPos := 1;
  FTextEndPos := Length(FText);
  FHideSelection := True;
  FAutoSelect := True;
  FAutoSize := True;

  FOnGetSymbolColor := nil;
  FOnAcceptChar := nil;
  FOldOnFontChange := Font.OnChange;
  Font.OnChange := FontChanged;

  SetSelection(0, 0);
  Cursor             := crIBeam;
  Width              := CE_DEF_WIDTH;
  Height             := CE_DEF_HEIGHT;
  TabStop            := CE_DEF_TABSTOP;
  ParentColor        := CE_DEF_PARENTCOLOR;

  ControlStyle := ControlStyle + [csClickEvents, csCaptureMouse, csOpaque];

  FTextColor := TSymbolColor.Create(clBlack, clWhite);
  FTextColor.OnColorChange := TextColorChanged;
  FSelectColor := TSymbolColor.Create(clWhite, clHighlight);
  FSelectColor.OnColorChange := SelectColorChanged; 
  FDisableColor := TSymbolColor.Create(clGray, clWhite);
  FDisableColor.OnColorChange := DisableColorChanged; 

  AdjustHeight;
  ClearUndoBuffer;
  ClearModifyState;
end;

// ---------------------------------------------------------------------------
// Destroy - 
// ---------------------------------------------------------------------------
destructor TColorEdit.Destroy;
begin
  FTextColor.Free;
  FSelectColor.Free;
  FDisableColor.Free;
  FCaret.Free;

  inherited Destroy;
end;

// ---------------------------------------------------------------------------
//  CreateParams -    -
//      
// ---------------------------------------------------------------------------
procedure TColorEdit.CreateParams(var Params : TCreateParams);
begin
  inherited CreateParams(Params);

  Params.Style := Params.Style or BorderStyles[FBorderStyle];

  if NewStyleControls and Ctl3D and (FBorderStyle = bsSingle) then begin
    Params.Style := Params.Style and not WS_BORDER;
    Params.ExStyle := Params.ExStyle or WS_EX_CLIENTEDGE;
  end;
end;

// ---------------------------------------------------------------------------
//  Paint -   !
// ---------------------------------------------------------------------------
procedure TColorEdit.Paint;
var
  I, CurXPos: Longint;
  hBmp  : HBITMAP;
  MemDC : HDC;
  TextSize: TSize;
  CR: TRect;
  cTextColor, cBkColor: TColor;
  LeftSelPos, RightSelPos: Longint;
  Select: Boolean;
begin
  //   ,     
  //   ActiveX!!!
  // if (Parent = nil) then Exit;

  //         
  //    !!!
  CalcTextEndPos;
  if FPaintLock <> 0 then Exit;

  //        (. PaintNeeded)
  ClearModifyState;

  //      ,    
  LeftSelPos  := Min(FSelBegPos, FSelEndPos);
  RightSelPos := Max(FSelBegPos, FSelEndPos);

  if Focused and Assigned(FCaret) then FCaret.Hide;

  //       
  //        ,  
  //         - API BitBlt
  CR := ClientRect;
  MemDC := CreateCompatibleDC(Canvas.Handle);
  hBmp := CreateCompatibleBitmap(Canvas.Handle, CR.Right, CR.Bottom);

  try
    SelectObject(MemDC, hBmp);
    SelectObject(MemDC, Canvas.Font.Handle);
    SetMapMode(MemDC, GetMapMode(Canvas.Handle));

    SetTextAlign(MemDC, TA_LEFT or TA_TOP);

    Canvas.Brush.Color := Color;
    FillRect(MemDC, CR, Canvas.Brush.Handle);
    CurXPos := LEFT_MARGIN;
    for I := FTextBegPos to FTextEndPos do begin
      if not Enabled then begin
        cTextColor := FDisableColor.TextColor;
        cBkColor   := FDisableColor.BackColor;
        Select := ((I >= LeftSelPos) and (I <= RightSelPos));
        if Assigned(FOnGetSymbolColor) then
          FOnGetSymbolColor(Self, FText[I], I, Select, cTextColor, cBkColor);
      end else if (esDisplaySelection in FEditState) and
                  (I >= LeftSelPos) and (I <= RightSelPos) then begin
        cTextColor := FSelectColor.TextColor;
        cBkColor   := FSelectColor.BackColor;
        if Assigned(FOnGetSymbolColor) then
          FOnGetSymbolColor(Self, FText[I], I, True, cTextColor, cBkColor);
      end else begin
        cTextColor := FTextColor.TextColor;;
        cBkColor   := FTextColor.BackColor;
        if Assigned(FOnGetSymbolColor) then
          FOnGetSymbolColor(Self, FText[I], I, False, cTextColor, cBkColor);
      end;
      Windows.SetTextColor(MemDC, ColorToRGB(cTextColor));
      Windows.SetBkColor(MemDC, ColorToRGB(cBkColor));
      ExtTextOut(MemDC, CurXPos, TOP_MARGIN, ETO_CLIPPED, @CR, @FText[I], 1, nil);
      GetTextExtentPoint32(MemDC, @FText[I], 1, TextSize);
      Inc(CurXPos, TextSize.cx);
    end;

    BitBlt(Canvas.Handle, 0, 0, CR.Right, CR.Bottom, MemDC, 0, 0, SrcCopy);
  finally
    DeleteDC(MemDC);
    DeleteObject(hBmp);
    if Focused and Assigned(FCaret) then FCaret.Show;
  end;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.CMCtl3DChanged(var Msg : TMessage);
begin
  if (csLoading in ComponentState) then
    Exit;
  if not HandleAllocated then
    Exit;

  if NewStyleControls and (FBorderStyle = bsSingle) then begin
    RecreateWnd;
    UpdateHeight;
  end;

  inherited;
end;

// ---------------------------------------------------------------------------
//  CMEnabledChanged
//    - Enabled  Paint -  
//     .
// (. RepaintWindow  ,     !!!)
// ---------------------------------------------------------------------------
procedure TColorEdit.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  Repaint;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.WMEraseBkGnd(var Msg : TWMEraseBkGnd);
begin
  Msg.Result := 1;  //      !
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.WMSetFont(var Msg : TWMSetFont);
begin
  inherited;
  //         - Font!!!
  Canvas.Font := Font; // ,    WM_PAINT!
  if Focused and Assigned(FCaret) then
    FCaret.Height := Canvas.TextHeight('Caret');
  AdjustText;
  AdjustHeight;
  if Focused then FCaret.Show;
end;

// ---------------------------------------------------------------------------
// SetText -   
// ---------------------------------------------------------------------------
procedure TColorEdit.SetText(Value: String);
begin
  if FText <> Value then begin
    SetSelection(0, 0);
    ClearUndoBuffer;
    FModified := False;
    if (FMaxLength > 0) and (FMaxLength < Length(Value)) then
      FText := Copy(Value, 1, FMaxLength)
    else
      FText := Value;
    FTextBegPos := 1;
    FCaretPos := FTextBegPos;

    RepaintWindow;
    PositionCaret;
  end;
end;

// ---------------------------------------------------------------------------
procedure  TColorEdit.Clear;
begin
  Text := '';
end;

// ---------------------------------------------------------------------------
//  WMSetFocus -    
// ---------------------------------------------------------------------------
procedure TColorEdit.WMSetFocus(var Msg : TWMSetFocus);
begin
  inherited;

  FCaret := TCaret.Create(Handle, 2*GetSystemMetrics(SM_CYBORDER),
                                  Canvas.TextHeight('Caret'));
  FCaret.SetPos(FCaretX, FCaretY);
  Include(FEditState, esDisplaySelection); //    

  // not (ssAlt in ShiftState) -  Alt    
  //   !!!
  if FAutoSelect and not (esMouseSelection in FEditState) and
     not (esRetainFocus in FEditState) and not (ssAlt in FShiftState) then
       SelectAll;

  Exclude(FEditState, esRetainFocus); //     

  Repaint;
  FCaret.Show;

  PostMessage(Handle, CEM_AFTERENTER, 0, 0); //    AfterEnter
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.WMKillFocus(var Msg : TWMKillFocus);
var
  NewWindow  : HWnd;
begin
  inherited;

  //   ,   
  if MouseCapture then MouseCapture := False;
  //    
  Exclude(FEditState, esMouseSelection);
  if FHideSelection then Exclude(FEditState, esDisplaySelection);

  FCaret.Free;
  FCaret := nil;

  //   -?
  NewWindow := Msg.FocusedWnd;
  if NewWindow = Handle then
    NewWindow := 0
  else if not IsSibling(NewWindow) then
    NewWindow := 0;

  //        -,   
  //        
  //  
  if NewWindow = 0 then Include(FEditState, esRetainFocus)
  else                  Exclude(FEditState, esRetainFocus);

  Repaint;

  PostMessage(Handle, CEM_AFTEREXIT, 0, 0); //    AfterEnter
end;

// ---------------------------------------------------------------------------
//  WMLButtonDown -    
//            
//   !
// ---------------------------------------------------------------------------
procedure TColorEdit.WMLButtonDown(var Msg : TWMLButtonDown);
var
  OldCaretPos: Longint;
begin
  inherited;

  if csDesigning in ComponentState then Exit;

  OldCaretPos := FCaretPos;
  CalcCaretPosOnMouseEvent(Msg.XPos, Msg.YPos);
  FCaret.SetPos(FCaretX, FCaretY);
  FShiftState := KeysToShiftState(Msg.Keys);
  if InSelectMode then begin
    PerformSelection(OldCaretPos);
    RepaintWindow;
  end else if IsSelected then begin
    SetSelection(0, 0);
    RepaintWindow;
  end;

  Include(FEditState, esMouseSelection);

  if not Focused then SetFocus;
end;

// ---------------------------------------------------------------------------
//  WMLButtonDblClk -     
// ---------------------------------------------------------------------------
procedure TColorEdit.WMLButtonDblClk(var Msg : TWMLButtonDblClk);
begin
  inherited;

  if csDesigning in ComponentState then Exit;
  CmdSelectWord;
end;

// ---------------------------------------------------------------------------
//  WMMouseMove -   ,     
// ---------------------------------------------------------------------------
procedure TColorEdit.WMMouseMove(var Msg : TWMMouseMove);
var
  OldCaretPos: Longint;
begin
  inherited;

  if csDesigning in ComponentState then
    Exit;
  if not Focused then
    Exit;

  if esMouseSelection in FEditState then begin
    if Msg.XPos < LEFT_MARGIN then begin
      CmdGoLeft;
    end else if Msg.XPos > RightBound then
      CmdGoRight
    else begin
      OldCaretPos := FCaretPos;
      CalcCaretPosOnMouseEvent(Msg.XPos, Msg.YPos);
      if FCaretPos <> OldCaretPos then begin
         PerformSelection(OldCaretPos);
         RepaintWindow;
         FCaret.SetPos(FCaretX, FCaretY);
         FCaret.Show;
      end;
    end;
  end;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.WMLButtonUp(var Msg : TWMLButtonUp);
begin
  inherited;
  //     
  Exclude(FEditState, esMouseSelection);
end;

// ---------------------------------------------------------------------------
//  WMRButtonUp -   
// ---------------------------------------------------------------------------
procedure TColorEdit.WMRButtonUp(var Msg : TWMRButtonDown);
var
  P  : TPoint;
  M  : TPopUpMenu;
  MI : TMenuItem;
begin
  if csDesigning in ComponentState then
    Exit;

  if not Focused and CanFocus then
    SetFocus;

  inherited;
  if PopUpMenu = nil then begin
    M := TPopupMenu.Create(Self);
    try
      MI := TMenuItem.Create(M);
      MI.Caption := MI_UNDO_TEXT;
      MI.Enabled := FUndoBuffer.Valid;
      MI.OnClick := UndoClick;
      M.Items.Add(MI);

      MI := TMenuItem.Create(M);
      MI.Caption := '-';
      M.Items.Add(MI);

      MI := TMenuItem.Create(M);
      MI.Caption := MI_CUT_TEXT;
      MI.Enabled := IsSelected and not FReadOnly;
      MI.OnClick := CutClick;
      M.Items.Add(MI);

      MI := TMenuItem.Create(M);
      MI.Caption := MI_COPY_TEXT;
      MI.Enabled := IsSelected;
      MI.OnClick := CopyClick;
      M.Items.Add(MI);

      MI := TMenuItem.Create(M);
      MI.Caption := MI_PASTE_TEXT;
      MI.Enabled := not FReadOnly and Clipboard.HasFormat(CF_TEXT);
      MI.OnClick := PasteClick;
      M.Items.Add(MI);

      MI := TMenuItem.Create(M);
      MI.Caption := MI_DELETE_TEXT;
      MI.Enabled := IsSelected and not FReadOnly;
      MI.OnClick := DeleteClick;
      M.Items.Add(MI);

      MI := TMenuItem.Create(M);
      MI.Caption := '-';
      M.Items.Add(MI);

      MI := TMenuItem.Create(M);
      MI.Caption := MI_SELECT_ALL_TEXT;
      MI.Enabled := not IsAllSelected;
      MI.OnClick := SelectAllClick;
      M.Items.Add(MI);

      P.X := Msg.XPos;
      P.Y := Msg.YPos;
      P := ClientToScreen(P);
      M.PopUp(P.X, P.Y);

      Application.ProcessMessages;
    finally
      M.Free;
    end;
  end;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.WMCopy(var Msg : TWMCopy);
begin
  inherited;
  CmdCopy;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.WMCut(var Msg : TWMCut);
begin
  inherited;
  CmdCut;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.WMPaste(var Msg : TWMPaste);
begin
  inherited;
  CmdPaste;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.WMClear(var Msg : TWMClear);
begin
  inherited;
  CmdDeleteSelected;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.CEMAfterEnter(var Msg : TMessage);
begin
  if Assigned(FAfterEnter) then
    FAfterEnter(Self);
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.CEMAfterExit(var Msg : TMessage);
begin
  if Assigned(FAfterExit) then
    FAfterExit(Self);
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.CopyToClipboard;
begin
  if HandleAllocated then
    Perform(WM_COPY, 0, 0);
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.CutToClipboard;
begin
  if HandleAllocated then
    Perform(WM_CUT, 0, 0);
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.PasteFromClipboard;
begin
  if HandleAllocated then
    Perform(WM_PASTE, 0, 0);
end;

// ---------------------------------------------------------------------------
//  GetSelText -   
// ---------------------------------------------------------------------------
function TColorEdit.GetSelText: String;
var
  BegPos, EndPos: Longint;
begin
  Result := '';
  if not IsSelected then Exit;

  BegPos := Min(FSelBegPos, FSelEndPos);
  EndPos := Max(FSelBegPos, FSelEndPos);

  Result := Copy(FText, BegPos, EndPos-BegPos+1);
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.CmdUndo;
var
  Buffer: TUndoRecord;
begin
  if not FUndoBuffer.Valid then Exit;
  Buffer := FUndoBuffer;

  SetUndoBuffer(CMD_UNDO);
  FText := Buffer.Text;
  with Buffer do begin
    FTextBegPos := TextBegPos;
    FTextEndPos := TextEndPos;
    FCaretPos   := CaretPos;
    FSelBegPos  := SelBegPos;
    FSelEndPos  := SelEndPos;
  end;

  RepaintWindow;
  PositionCaret;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.CmdCopy;
var
  H  : THandle;
  S  : String;
  GP : PAnsiChar;
  Len: Longint;
begin
  S := GetSelText;
  if S = '' then Exit;

  Len := Length(S);
  H := GlobalAlloc(GHND, Len+1);

  if H = 0 then Exit;

  GP := GlobalLock(H);
  try
    Move(S[1], GP^, Len);
    Inc(GP, Len);
    GP^ := #0;
  finally
    GlobalUnlock(H);
  end;

  OpenClipboard(Handle);
  try
    EmptyClipboard;
    SetClipboardData(CF_TEXT, H);
  finally
    CloseClipboard;
  end;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.CmdPaste;
var
  H   : THandle;
  P   : PAnsiChar;
  S: String;
  I, BegCaretPos: Longint;
begin
  if FReadOnly then Exit;

  //         CmdDeleteSelected!
  PaintLock;
  FCaret.Hide;

  OpenClipboard(Handle);
  try
    if IsSelected then begin
      CmdDeleteSelected;
      SetUndoCommand(CMD_CHAR);
    end;

    SetUndoBuffer(CMD_CHAR);

    H := GetClipboardData(CF_TEXT);
    if H <> 0 then begin
      P := PAnsiChar(GlobalLock(H));
      try
        SetString(S, P, StrLen(P));

        BegCaretPos := FCaretPos;
        //    - CmdChar    !!!
        for I := 1 to Length(S) do
          CmdChar(Word(S[I]));

        FSelBegPos := BegCaretPos;
        if FCaretPos > 1 then FSelEndPos := FCaretPos - 1
        else                  FSelEndPos := FSelBegPos;
      finally
        GlobalUnlock(H);
      end;
    end;
  finally
    CloseClipboard;
    PaintUnlock;
    FCaret.Show;
  end;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.CmdCut;
begin
  if FReadOnly then Exit;

  SetUndoBuffer(CMD_CUT);
  CmdCopy;
  CmdDeleteSelected;
end;

// ---------------------------------------------------------------------------
//  CmdSelectWord -    
// ---------------------------------------------------------------------------
procedure TColorEdit.CmdSelectWord;
var
  WordBegPos, WordEndPos: Longint;
  Len: Longint;
begin
  Len := Length(FText);
  if Len = 0 then
    Exit;

  WordBegPos := FCaretPos-1;
  if WordBegPos = 0 then Inc(WordBegPos);

  while WordBegPos > 1 do begin
    if not (FText[WordBegPos] in EMPTY_CHAR) then
      break;
    Dec(WordBegPos);
  end;
  while WordBegPos > 1 do begin
    if FText[WordBegPos-1] in EMPTY_CHAR then
      break;
    Dec(WordBegPos);
  end;

  WordEndPos := FCaretPos-1;
  if WordEndPos = 0 then Inc(WordEndPos);
  while WordEndPos < Len do begin
    if FText[WordEndPos] in EMPTY_CHAR then
      break;
    Inc(WordEndPos);
  end;
  while WordBegPos < Len do begin
    if not (FText[WordEndPos+1] in EMPTY_CHAR) then
      break;
    Inc(WordEndPos);
  end;

  FCaretPos  := WordEndPos + 1;
  if FTextEndPos < WordEndPos then begin
    FTextEndPos := WordEndPos;
    CalcTextBegPos;
  end;
  FSelBegPos := WordBegPos;
  FSelEndPos := WordEndPos;

  RepaintWindow;
  PositionCaret;
end;

// ---------------------------------------------------------------------------
//  CmdChar -      
// ---------------------------------------------------------------------------
procedure TColorEdit.CmdChar(CharCode: Word);
var
  InsStr: String;
  InsChr: Char;
  Accept: Boolean;
begin
  if FReadOnly then Exit;

  if CharCode in [0, VK_BACK, VK_ESCAPE] then Exit;

  if not (Char(CharCode) in [#32..#255]) then begin
    MessageBeep(Cardinal(-1));
    Exit;
  end;

  if (FMaxLength > 0) and (Length(FText) >= FMaxLength) then
    Exit;

  InsChr := Char(CharCode);
  SetString(InsStr, PChar(@InsChr), 1);

  case FCharCase of
    ecUpperCase: InsStr := AnsiUpperCase(InsStr);
    ecLowerCase: InsStr := AnsiLowerCase(InsStr);
  end;

  Accept := True;
  if Assigned(FOnAcceptChar) then
    FOnAcceptChar(Self, InsStr[1], Accept);

  if not Accept then Exit;

  PaintLock; //     CmdDeleteSelected!
  try
    if IsSelected then begin
      CmdDeleteSelected;
      SetUndoCommand(CMD_CHAR);
    end;

    SetUndoBuffer(CMD_CHAR);

    Insert(InsStr, FText, FCaretPos);

    ClearSelectMode;
    CalcTextEndPos;
    CmdGoRight;
  finally
    PaintUnlock;
  end;
end;

// ---------------------------------------------------------------------------
//  CmdDeleteSelected -   
// ---------------------------------------------------------------------------
procedure TColorEdit.CmdDeleteSelected;
var
  LeftSelPos, RightSelPos: Longint;
  UndoValid: Boolean;
begin
  if not IsSelected or FReadOnly then Exit;

  SetUndoBuffer(CMD_CUT);

  if IsAllSelected then begin //  
    // Clear   FUndoBuffer.Valid,    !
    UndoValid := FUndoBuffer.Valid;
    Clear;
    FUndoBuffer.Valid := UndoValid;
    Exit;
  end;

  if FSelEndPos >  FSelBegPos then begin
    LeftSelPos  := FSelBegPos;
    RightSelPos := FSelEndPos;
  end else begin
    LeftSelPos  := FSelEndPos;
    RightSelPos := FSelBegPos;
  end;

  Delete (FText, LeftSelPos, RightSelPos - LeftSelPos + 1);
  if FSelEndPos >= FSelBegPos then Dec(FCaretPos, RightSelPos - LeftSelPos + 1);
  SetSelection(0, 0);

  RepaintWindow;
  PositionCaret;
end;

// ---------------------------------------------------------------------------
//  CmdDeleteChar -   /  
// ---------------------------------------------------------------------------
procedure TColorEdit.CmdDeleteChar(Dir: Longint);
begin
  if FReadOnly then Exit;

  if Dir < 0 then begin
    if FCaretPos = 1 then Exit;
    SetUndoBuffer(CMD_CUT);
    Delete(FText, FCaretPos-1, 1);
    RepaintWindow;
    CmdGoLeft;
  end else begin
    if FCaretPos = Length(FText) + 1 then Exit;
    SetUndoBuffer(CMD_CUT);
    Delete(FText, FCaretPos, 1);
    RepaintWindow;
  end;
end;

// ---------------------------------------------------------------------------
//  CmdGoLeft -    
//     ,    
// ---------------------------------------------------------------------------
procedure TColorEdit.CmdGoLeft;
var
  OldCaretPos: Longint;
begin
  //        (FSelBegPos, FSelEndPos)
  CheckSelectMode;

  OldCaretPos := FCaretPos;
  if (FTextBegPos > 1) and (FCaretPos <= FTextBegPos + 1) then begin
    ScrollText(TEXT_SCROLL_CHUNK);
    Dec(FCaretPos);
    if not InSelectMode then RepaintWindow;
  end else if FCaretPos > FTextBegPos then begin
    Dec(FCaretPos);
  end;

  if InSelectMode then  //  
    PerformSelection(OldCaretPos);

  RepaintWindow;
  if OldCaretPos <> FCaretPos then
    PositionCaret;
end;

// ---------------------------------------------------------------------------
//  CmdGoRigh -    
//     ,    
// ---------------------------------------------------------------------------
procedure TColorEdit.CmdGoRight;
var
  OldCaretPos: Longint;
begin
  //        (FSelBegPos, FSelEndPos)
  CheckSelectMode;

  OldCaretPos := FCaretPos;
  if (FTextEndPos < Length(FText)) and (FCaretPos >= FTextEndPos + 1) then begin
    //  .     !!
    ScrollText(-TEXT_SCROLL_CHUNK);
    Inc(FCaretPos);
    if not InSelectMode then RepaintWindow;
  end else if (FCaretPos <= FTextEndPos) then begin
    Inc(FCaretPos);
  end;

  if InSelectMode then //  
    PerformSelection(OldCaretPos);

  RepaintWindow;
  if OldCaretPos <> FCaretPos then
    PositionCaret;
end;

// ---------------------------------------------------------------------------
//  CmdGoHome -    
//     ,    
// ---------------------------------------------------------------------------
procedure TColorEdit.CmdGoHome;
var
  OldCaretPos: Longint;
begin
  CheckSelectMode;

  OldCaretPos := FCaretPos;
  FTextBegPos := 1;
  FCaretPos := FTextBegPos;

  if InSelectMode then
    PerformSelection(OldCaretPos);

  RepaintWindow;
  if OldCaretPos <> FCaretPos then
    PositionCaret;
end;

// ---------------------------------------------------------------------------
//  CmdGoEnd -    
//     ,    
// ---------------------------------------------------------------------------
procedure TColorEdit.CmdGoEnd;
var
  OldCaretPos: Longint;
begin
  CheckSelectMode;

  OldCaretPos := FCaretPos;
  FTextEndPos := Length(FText);
  FCaretPos := FTextEndPos + 1;
  CalcTextBegPos;

  if InSelectMode then
    PerformSelection(OldCaretPos);

  RepaintWindow;
  if OldCaretPos <> FCaretPos then
    PositionCaret;
end;

// ---------------------------------------------------------------------------
//  CalcCaretPosOnMouseEvent -      
//                              
// ---------------------------------------------------------------------------
procedure TColorEdit.CalcCaretPosOnMouseEvent(X, Y: Longint);
var
  I, TextX, LastX: Longint;
begin
  FCaretX := LEFT_MARGIN +
             Canvas.TextWidth(Copy(FText, FTextBegPos, FTextEndPos - FTextBegPos + 1));
  FCaretPos := FTextEndPos + 1;

  LastX := LEFT_MARGIN;
  TextX := LEFT_MARGIN;
  for I := FTextBegPos to FTextEndPos do begin
    TextX := TextX + Canvas.TextWidth(FText[I]);
    if TextX > X then begin
      if (X - LastX) > ((TextX - LastX) div 2) then begin
        FCaretX := TextX;
        FCaretPos := I + 1;
      end else begin
        FCaretX := LastX;
        FCaretPos := I;
      end;
      break;
    end;
    LastX := TextX;
  end;
end;

// ---------------------------------------------------------------------------
//  WMSysKeyDown -     <Alt+BakSpace>
// ---------------------------------------------------------------------------
procedure TColorEdit.WMSysKeyDown(var Msg : TWMSysKeyDown);
begin
  inherited;

  if (csDesigning in ComponentState) or FReadOnly then
    Exit;
  FShiftState := KeyDataToShiftState(Msg.KeyData);
  if (Msg.CharCode = VK_BACK) and (ssAlt in FShiftState) then
     CmdUndo;
end;

// ---------------------------------------------------------------------------
// WMKeyDown -   
// ---------------------------------------------------------------------------
procedure TColorEdit.WMKeyDown(var Msg : TWMKeyDown);
begin
  inherited;

  if (csDesigning in ComponentState) then
    Exit;

  FShiftState := KeyDataToShiftState(Msg.KeyData);

  case Msg.CharCode of
    VK_LEFT:    begin
      CmdGoLeft;
    end;
    VK_RIGHT:   begin
      CmdGoRight;
    end;
    VK_HOME:    begin
      CmdGoHome;
    end;
    VK_END:     begin
      CmdGoEnd;
    end;
    VK_DELETE:  begin
      if IsSelected then  CmdDeleteSelected
      else                CmdDeleteChar(1);
    end;
    VK_BACK: begin
      if IsSelected then CmdDeleteSelected
      else               CmdDeleteChar(-1);
    end;
    VK_INSERT: begin
      if      ssCtrl  in FShiftState then CopyToClipboard
      else if ssShift in FShiftState then PasteFromClipboard;
    end;
    VK_CLEAR: Clear;
  end;
end;

// ---------------------------------------------------------------------------
// WMChar -  
// ---------------------------------------------------------------------------
procedure TColorEdit.WMChar(var Msg : TWMChar);
begin
  inherited;

  if (csDesigning in ComponentState) then
    Exit;

  //  /
  if ssCtrl in FShiftState then begin
    case Msg.CharCode of
      CMD_COPY : CopyToClipboard;
      CMD_CUT  : CutToClipboard;
      CMD_PASTE: PasteFromClipboard;
      CMD_UNDO : CmdUndo;
    end;
    Exit;
  end;

  CmdChar(Msg.CharCode);
end;

// ---------------------------------------------------------------------------
//  WMGetDlgCode -   !!!
//       ,   <Left. Right, Up, Down>
//        -  ,   
//    <Tab>!!!
// ---------------------------------------------------------------------------
procedure TColorEdit.WMGetDlgCode(var Msg : TWMGetDlgCode);
begin
  inherited;

  if csDesigning in ComponentState then
    Msg.Result := DLGC_STATIC
  else
    Msg.Result := Msg.Result or DLGC_WANTCHARS or DLGC_WANTARROWS;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.UndoClick(Sender : TObject);
begin
  CmdUndo;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.CutClick(Sender : TObject);
begin
  CutToClipboard;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.CopyClick(Sender : TObject);
begin
  CopyToClipboard;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.PasteClick(Sender : TObject);
begin
  PasteFromClipboard;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.DeleteClick(Sender : TObject);
begin
  CmdDeleteSelected;
end;

// ---------------------------------------------------------------------------
procedure TColorEdit.SelectAllClick(Sender : TObject);
begin
  SelectAll;
end;

// ---------------------------------------------------------------------------

end. { ColorEdit }
