//==========================================================================//
//                                                                          //
// TColorMemo v1.20 (C)  .. 1998                                    //
//                                                                          //
//==========================================================================//

unit ColorMem;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, Buttons, Caret, Format, SymColor;

//
//   
//
//{$DEFINE DELPHI_400}
//{$DEFINE DEBUG}
//{$DEFINE DEBUG_TEXT_STATISTIC}
//{$DEFINE DEBUG_SELECTION}
//{$DEFINE DEBUG_CARET}
//{$DEFINE DEBUG_FORMAT_LINES}
//{$DEFINE DEBUG_SCROLL_RANGE}

const
  //  TColorMemo
  CMM_AFTERENTER = WM_USER + 100;
  CMM_AFTEREXIT  = WM_USER + 101;

type
  //       
  TMemoColorEvent = procedure (Sender: TObject; Ch: Char;
                           Row, Col: Longint; Selected: Boolean;
                           var TextColor, BkColor: TColor) of object;

  //     
  //    .  (. FormatText, GetPaintInfo)
  TFormatInfo = record
    Valid: Boolean;      // False -  
    BegIndex: Longint;   //  
    EndIndex: Longint;   //  
  end;

  //      (. GetPaintInfo)
  TPaintInfo = record
    BegR, EndR: Longint;          // .  .  . 
    TopY: Longint;                // Y-. .   Rect
    Rect: TRect;                  //  
    TopFlag, BottomFlag: Boolean; // + ,  ?
  end;

  //   
  TTextStatistic = record
    TotalWidth: Longint;     // .  
    AvgTailCharW: Double;    //   ,   
    TailCharCount: Longint;  // - ,   
    VisibleRows: Longint;    // -    
    VertPageSize: Longint;   // . - . 
    VisibleChars: Longint;   // -    
  end;

  //    ScrollBars
  TScrollState = record
    VertMinPos, VertMaxPos: Longint;
    VertPos: Longint;
    HorMinPos, HorMaxPos: Longint;
    HorPos: Longint;
  end;

  //  
  //  0..Count
  //  1..Length(String)+1
  TCaretState = record
    Row , Col : Longint; //     . 
    X, Y: Longint;       //   
  end;

  //      
  // Col    0   !!!
  TSelectState = record
    RowBeg, ColBeg: Longint; // 
    RowEnd, ColEnd: Longint; // 
  end;

  //   
  //     
  TModifyState = record
    FormatLines: TFormatter;   //  
    SelectState: TSelectState; //   
    ScrollState: TScrollState; //  
    Text: String;              // 
  end;

  //      
  //    , 
  //        ,
  //         (. SetUndoBuffer)
  TUndoRecord = record
    Valid: Boolean;              // True -  
    LastCommand: Longint;        //   
                                 // .  
    Text: String;                // 
    CaretState: TCaretState;     //  
    SelectState: TSelectState;   //   
    ScrollState: TScrollState;   //  
  end;

  //  
  //------------------------------------------------------------
  // msInvFormat       -   
  // msInvScrollRange  -    
  // msInvScrollPos    -   
  // msMouseSelection  -     
  // msTextChanging    -     Memo
  // msCaretOutOfRange -     
  //                       ,  / 
  //                       
  // msRepaintWindow   - ,  ,   
  //                   -  Paint:  RepaintWindow  
  // msPaintAll        -    Memo

  TMemoState = (msInvFormat, msInvScrollRange, msInvScrollPos, msMouseSelection,
                msTextChanging, msCaretOutOfRange, msRepaintWindow, msPaintAll);
  TMemoStates = set of TMemoState;


  {TColorMemo}

  TColorMemo = class(TCustomControl)
  private
    FFormatLines: TFormatter;
    FCaret: TCaret;
    FMemoState: TMemoStates;
    FCaretState: TCaretState;
    FScrollState: TScrollState;
    FSelectState: TSelectState;
    FShiftState: TShiftState;
    FTextStatistic: TTextStatistic;
    FPaintLock: Integer;              //   
    FModifyState: TModifyState;
    FFormatInfo: TFormatInfo;
    FPaintInfo: TPaintInfo;
    FUndoBuffer: TUndoRecord;

    FLines: TStrings;                 //  
    FBorderStyle: TBorderStyle;
    FScrollBars: TScrollStyle;
    FReadOnly: Boolean;
    FWordWrap: Boolean;
    FHideSelection: Boolean;
    FWantReturns: Boolean;
    FMaxLength: Integer;               //   
                                       // (  )
    FTextColor   : TSymbolColor;       //  
    FSelectColor : TSymbolColor;       //   
    FDisableColor: TSymbolColor;       //    Enabled := False
    FModified: Boolean;                //   ""

    //   
    FOnChange   : TNotifyEvent;         //   
    FOnStartPaint: TNotifyEvent;        //  
    FOnEndPaint: TNotifyEvent;          //  
    FOnGetSymbolColor: TMemoColorEvent; //  .    
    FAfterEnter : TNotifyEvent;         //    
    FAfterExit  : TNotifyEvent;         //    

    //       ,
    //  TControl,       ,
    //    - TextColor  Font
    FOldOnFontChange: TNotifyEvent;

    procedure SetUndoCommand(Command: Longint);
    procedure ClearUndoBuffer;
    procedure SetUndoBuffer(Command: Longint);

    //  -
    {$IFNDEF DELPHI_400}
    function StrToOem(S: String): String;
    {$ENDIF}
    procedure DebugLn(S: String);
    {$IFDEF DEBUG}
    procedure DebugSelection;
    procedure DebugSelectionReal;
    procedure DebugCaretPos;
    procedure DebugTextStatistic;
    procedure DebugFormatLines;
    procedure DebugModifyFormatLines;
    procedure DebugScrollRange;
    {$ENDIF}

    //  Clipboard
    procedure CmdCopy;
    procedure CmdCut;
    procedure CmdPaste;
    //  
    procedure CmdUndo;
    procedure CmdChar(CharCode: Word);
    //   Dir < 0 -   ,  Dir > 0 -   
    procedure CmdDeleteChar(Dir: Longint);
    procedure CmdNewLine;
    procedure CmdCombineLines(StartRow, EndRow: Longint);
    procedure CmdDeleteSelected;
    //   
    procedure CmdGoLeft;
    procedure CmdGoRight;
    procedure CmdGoUp;
    procedure CmdGoPgUp;
    procedure CmdGoDown;
    procedure CmdGoPgDown;
    procedure CmdGoHome;
    procedure CmdGoEnd;
    procedure CmdGoStart;
    procedure CmdGoFinish;
    procedure CmdSelectWord;
    //  
    function  IsChanged: Boolean;
    procedure InsertFirstLine;
    function  CreateFormatLines: TFormatter;
    procedure ModifyLine(Index: Longint; NewLine: String);
    procedure GetPaintInfo;
    procedure CaretPosRealToFmt(RealRow, RealCol: Longint;
                                var FmtRow, FmtCol: Longint);
    procedure CaretPosFmtToReal(var RealRow, RealCol: Longint;
                                FmtRow, FmtCol: Longint);
    function  GetCaretHeight: Longint;
    procedure RepaintWindow;
    function  PaintNeeded: Boolean;
    procedure ClearCaretState;
    procedure ClearScrollState;
    procedure ClearModifyState;
    procedure PaintLock;
    procedure PaintUnlock;
    procedure PerformSelection(OldCaretRow, OldCaretCol: Longint);
    procedure GetSelMinMaxInfo(var RowMin, RowMax, ColMin, ColMax: Longint);
    function  GetSelText: String;
    function  IsSelectedChar(Row, Col: Longint): Boolean;
    function  InSelectMode: Boolean;
    procedure CheckSelectMode;
    function  IsSelected: Boolean;
    function  IsAllSelected: Boolean;
    procedure ResetSelection;
    procedure MakeRowBegVisible;
    procedure MakeCaretVisible;
    procedure MakeRowEndVisible;
    function  GetFirstRow: Longint;
    procedure ScrollVert(Pos: Smallint);
    procedure ScrollHor(Pos: Smallint);
    function  GetFormatLine(FmtRow: Longint): String;
    function  GetPaintBufferWidth: Longint;
    function  GetHorOffset: Longint;
    procedure UpdateCaretYPos;
    procedure UpdateCaretXPos;
    function  CaretIsVertVisible: Boolean;
    function  CaretIsHorVisible: Boolean;
    function  RowIsVisible(FmtRow: Longint): Boolean;
    function  GetRowEndPos(FmtRow: Longint; var CaretX: Longint): Longint;
    procedure PositionCaretXY(X, Y: Longint);
    procedure SetLines(Value: TStrings);
    procedure SetScrollBars(Value: TScrollStyle);
    procedure SetBorderStyle(Value: TBorderStyle);
    procedure SetWordWrap(Value: Boolean);
    procedure SetMaxLength(Value: Integer);
    procedure SetTextColor(Value: TSymbolColor);
    procedure SetSelectColor(Value: TSymbolColor);
    procedure SetDisableColor(Value: TSymbolColor);
    function  GetSelStart: Longint;
    procedure SetSelStart(Value: Longint);
    function  GetSelLength: Longint;
    procedure SetSelLength(Value: Longint);
    procedure SetHideSelection(Value: Boolean);
    procedure SetVersion(Value: String);
    function  GetModified: Boolean;
    procedure SetModified(Value: Boolean);
    procedure UpdateScrollRange;
    procedure ResetScrollPos;
    procedure UpdateScrollPos;
    procedure GetTextStatistic;
    procedure FormatText;
    procedure InvalidateState;
    procedure TextChanged(Sender: TObject);
    procedure UpdateAll(ResetScroll: Boolean);
    procedure FontChanged(Sender: TObject);
    procedure TextColorChanged(Sender: TObject);
    procedure SelectColorChanged(Sender: TObject);
    procedure DisableColorChanged(Sender: TObject);
    procedure ReplaceSelection(NewSel: String);
    procedure ClearState;
    function  GetText: String;
    procedure SetText(Value: String);
  protected
    FVersion: String; //  

    procedure CreateParams(var Params : TCreateParams); override;
    procedure Paint; override;
    procedure Loaded; override;
    //     ,  
    //      !
    procedure Change; virtual;
    //   
    procedure UndoClick(Sender : TObject);      dynamic;
    procedure CutClick(Sender : TObject);       dynamic;
    procedure CopyClick(Sender : TObject);      dynamic;
    procedure PasteClick(Sender : TObject);     dynamic;
    procedure DeleteClick(Sender : TObject);    dynamic;
    procedure SelectAllClick(Sender : TObject); dynamic;
    //  Windows
    procedure WMVScroll(var Msg: TWMVScroll); message WM_VSCROLL;
    procedure WMHScroll(var Msg: TWMHScroll); message WM_HSCROLL;
    procedure WMSize(var Msg: TWMSize);
    procedure WMEraseBkGnd(var Msg : TWMEraseBkGnd); message WM_ERASEBKGND;
    procedure WMWindowPosChanged(var Message: TWMWindowPosChanged);
                                                   message WM_WINDOWPOSCHANGED;
    procedure WMSetFocus(var Msg : TWMSetFocus); message WM_SETFOCUS;
    procedure WMKillFocus(var Msg : TWMKillFocus); message WM_KILLFOCUS;
    procedure WMLButtonDown(var Msg : TWMLButtonDown); message WM_LBUTTONDOWN;
    procedure WMLButtonUp(var Msg : TWMLButtonUp); message WM_LBUTTONUP;
    procedure WMMouseMove(var Msg : TWMMouseMove); message WM_MOUSEMOVE;
    procedure WMGetDlgCode(var Msg : TWMGetDlgCode); message WM_GETDLGCODE;
    procedure WMKeyDown(var Msg : TWMKeyDown); message WM_KEYDOWN;
    procedure WMSysKeyDown(var Msg : TWMSysKeyDown); message WM_SYSKEYDOWN;
    procedure WMCopy(var Msg : TWMCopy); message WM_COPY;
    procedure WMCut(var Msg : TWMCut); message WM_CUT;
    procedure WMPaste(var Msg : TWMPaste); message WM_PASTE;
    procedure WMSetFont(var Msg: TWMSetFont); message WM_SETFONT;
    // procedure WMClear(var Msg : TWMClear); message WM_CLEAR;
    procedure WMLButtonDblClk(var Msg : TWMLButtonDblClk); message WM_LBUTTONDBLCLK;
    procedure WMChar(var Msg : TWMChar); message WM_CHAR;
    procedure WMRButtonUp(var Msg : TWMRButtonDown); message WM_RBUTTONUP;
    //  Delphi
    procedure CMCtl3DChanged(var Msg : TMessage); message CM_CTL3DCHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    //   
    procedure CMMAfterEnter(var Msg : TMessage); message CMM_AFTERENTER;
    procedure CMMAfterExit(var Msg : TMessage);  message CMM_AFTEREXIT;

  public
    //   
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    procedure   CopyToClipboard;
    procedure   CutToClipboard;
    procedure   PasteFromClipboard;
    procedure   SelectAll;
    procedure   Clear;
    procedure   ClearSelection;

    //    1.10
    procedure ClearUndo;
    procedure Undo;

    //  ? (. TEdit)
    property Modified: Boolean
             read GetModified     write SetModified;
    property SelStart: Longint
             read GetSelStart     write SetSelStart;
    property SelLength: Longint
             read GetSelLength    write SetSelLength;
    property SelText: String
             read GetSelText      write ReplaceSelection;
published
    //  (design-time)
    property BorderStyle: TBorderStyle
             read FBorderStyle    write SetBorderStyle default bsSingle;
    property ScrollBars: TScrollStyle
             read FScrollBars     write SetScrollBars default ssNone;
    property Lines: TStrings
             read FLines          write SetLines;
    property ReadOnly: Boolean
             read FReadOnly       write FReadOnly default False;
    property WordWrap: Boolean
             read FWordWrap       write SetWordWrap default True;
    property HideSelection: Boolean
             read FHideSelection  write SetHideSelection default True;
    property WantReturns: Boolean
             read FWantReturns    write FWantReturns default True;
    property MaxLength: Integer
             read FMaxLength      write SetMaxLength   default 0;
    property TextColor: TSymbolColor
             read FTextColor      write SetTextColor;
    property SelectColor: TSymbolColor
             read FSelectColor    write SetSelectColor;
    property DisableColor: TSymbolColor
             read FDisableColor   write SetDisableColor;
    property Text: String
             read GetText         write SetText;
    property Version: String
             read FVersion        write SetVersion stored False;

    // -   TCustomControl (design-time)
    property Align;
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;

    //    1.10
    property Anchors;
    property Constraints;
    property DragKind;

    //  (design-time)
    property OnGetSymbolColor: TMemoColorEvent
             read FOnGetSymbolColor     write FOnGetSymbolColor;
    property OnChange: TNotifyEvent
             read FOnChange       write FOnChange;
    property OnStartPaint: TNotifyEvent
             read FOnStartPaint   write FOnStartPaint;
    property OnEndPaint: TNotifyEvent
             read FOnEndPaint     write FOnEndPaint;
    property AfterEnter: TNotifyEvent
             read FAfterEnter     write FAfterEnter;
    property AfterExit: TNotifyEvent
             read FAfterExit      write FAfterExit;

    // ,   TCustomControl (design-time)
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;

    //    1.10
    property OnEndDock;
    property OnStartDock;
  end;
  {  TColorMemo}

implementation

uses {$IFDEF DELPHI_400} StrUtils,{$ENDIF} Clipbrd, Menus;

function Min (A, B: Longint): Longint;
begin
  Result := A;
  if B < A then Result := B;
end;

function Max (A, B: Longint): Longint;
begin
  Result := A;
  if B > A then Result := B;
end;


{TColorMemo}

const
  BorderStyles: array[TBorderStyle] of Longint = (0, WS_BORDER);
  LEFT_MARGIN        = 2;        //   
  RIGHT_MARGIN       = 2;        //   
  TOP_MARGIN         = 2;        //   
  BOTTOM_MARGIN      = 2;        //   
  HOR_SCROLL_CHUNK   = 3;        //    ( )
  VERT_SCROLL_CHUNK  = 3;        //    ( )
  CM_DEF_WIDTH       = 185;      //  -
  CM_DEF_HEIGHT      = 89;       //  -
  CM_DEF_PARENTCOLOR = False;    //   -
  MAX_STR_LEN        = 1024;     // .   ( )

  // .    CMD_CHAR      
  CMD_NULL   =    0;
  CMD_COPY   =    3;
  CMD_PASTE  =   22;
  CMD_CUT    =   24;
  CMD_UNDO   =   26;
  CMD_CHAR   = 1001;

  CRLF = #13#10;  //  ,  

  //   
  MI_UNDO_TEXT        = '&';
  MI_CUT_TEXT         = '&';
  MI_COPY_TEXT        = '&';
  MI_PASTE_TEXT       = '&';
  MI_DELETE_TEXT      = '&';
  MI_SELECT_ALL_TEXT  = ' &';


  //  
  COLOR_MEMO_VERSION = '1.20';

//----------------------------------------------------------------------------
//    1.10
//----------------------------------------------------------------------------
procedure TColorMemo.ClearUndo;
begin
  ClearUndoBuffer;
end;

procedure TColorMemo.Undo;
begin
  CmdUndo;
end;

// ---------------------------------------------------------------------------
// GetText -   .
//   Lines.Text   CRLF
// ---------------------------------------------------------------------------
function  TColorMemo.GetText: String;
var
  Len: Longint;
begin
  Result := FLines.Text;
  Len := Length(Result);
  if (Len >= 2) and (Result[Len-1] = #13) and (Result[Len] = #10) then
    Delete(Result, Len-1, 2);
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.SetText(Value: String);
begin
  if Value <> GetText then
    FLines.Text := Value;
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.SetHideSelection(Value: Boolean);
begin
  if Value <> FHideSelection then begin
    FHideSelection := Value;
    Repaint;
  end;
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.ClearSelection;
begin
  ResetSelection;
end;

// ---------------------------------------------------------------------------
// GetSelStart -  .   
// ---------------------------------------------------------------------------
function  TColorMemo.GetSelStart: Longint;
var
  I: Longint;
  RowMin, RowMax, ColMin, ColMax: Longint;
begin
  Result := 0;

  if IsSelected then begin
    GetSelMinMaxInfo(RowMin, RowMax, ColMin, ColMax);
    with FFormatLines do
      for I := 0 to RowMin-1 do
        Result := Result + Length(GetFormatLine(I));
    Inc(Result, ColMin);
  end else begin // 1.10
    with FFormatLines do
      for I := 0 to FCaretState.Row-1 do
        Result := Result + Length(GetFormatLine(I));
    Inc(Result, FCaretState.Col);
  end;
end;

// ---------------------------------------------------------------------------
// SetSelStart -  .   
//       
// ---------------------------------------------------------------------------
procedure TColorMemo.SetSelStart(Value: Longint);
var
  I, P, Len: Longint;
begin
  if (Value < 0) or (FLines.Count = 0) then Exit;
  ResetSelection;

  // 1.11 ()
  Len := 0;
  //
  with FCaretState do begin
    Row := 0;
    P := 0;
    for I := 0 to FFormatLines.Count-1 do begin
      Row := I;
      Len := Length(GetFormatLine(I));
      if P + Len + 1 >= Value then
        break;
      Inc(P, Len);
    end;

    // 1.10 ()
    Col := Min(Len+1, Value - P);
    //

    UpdateCaretXPos;
    UpdateCaretYPos;
  end;
  Repaint;
end;

// ---------------------------------------------------------------------------
// GetSelLength -    
// ---------------------------------------------------------------------------
function  TColorMemo.GetSelLength: Longint;
var
  I: Longint;
  RowMin, RowMax, ColMin, ColMax: Longint;
begin
  Result := 0;
  if (FLines.Count = 0) or not IsSelected then Exit;

  GetSelMinMaxInfo(RowMin, RowMax, ColMin, ColMax);
  if RowMax > RowMin then begin
    Result := Length(GetFormatLine(RowMin)) - ColMin + 1;
    for I := RowMin+1 to RowMax-1 do
      Result := Result + Length(GetFormatLine(I));
    Inc(Result, ColMax);
  end else
   Result := ColMax - ColMin + 1;
end;

// ---------------------------------------------------------------------------
// SetSelLength -    
// ---------------------------------------------------------------------------
procedure TColorMemo.SetSelLength(Value: Longint);
var
  OldCaretRow, OldCaretCol, RowBefore: Longint;
  I: Longint;
begin
  if Value = GetSelLength then Exit;
  ResetSelection;

  with FCaretState do begin
    OldCaretRow := Row;
    OldCaretCol := Col;
  end;

  if FCaret <> nil then
    FCaret.Hide;
  PaintLock;
  Include(FMemoState, msMouseSelection);
  try
    I := 0;
    while I < Abs(Value) do begin
      RowBefore := FCaretState.Row;
      if Value > 0 then CmdGoRight
      else              CmdGoLeft;
      if RowBefore = FCaretState.Row then
        Inc(I);
    end;
    PerformSelection(OldCaretRow, OldCaretCol);
  finally
    Exclude(FMemoState, msMouseSelection);
    PaintUnlock;
    if FCaret <> nil then
      FCaret.Show;
  end;
end;

// ---------------------------------------------------------------------------
// ReplaceSelection -    
// ---------------------------------------------------------------------------
procedure TColorMemo.ReplaceSelection(NewSel: String);
var
  I: Longint;
begin
  PaintLock;
  try
    CmdDeleteSelected;
    for I := 1 to Length(NewSel) do
      CmdChar(Word(NewSel[I]));
  finally
    PaintUnlock;
  end;
end;

// ---------------------------------------------------------------------------
// Clear -   
// ---------------------------------------------------------------------------
procedure  TColorMemo.Clear;
begin
  FLines.Clear;
end;

// ---------------------------------------------------------------------------
// GetModified -  ?
//     : FUndoBuffer.Valid
//                                "" : FModified
// ---------------------------------------------------------------------------
function  TColorMemo.GetModified: Boolean;
begin
  Result := FModified or FUndoBuffer.Valid;
end;

// ---------------------------------------------------------------------------
// SetModified -   ""    Change
//  False,   UndoBuffer
// ---------------------------------------------------------------------------
procedure TColorMemo.SetModified(Value: Boolean);
begin
  if Value <> FModified then begin
    FModified := Value;
    if not Value then ClearUndoBuffer;
    Change;
  end;
end;

// ---------------------------------------------------------------------------
//  Loaded -       OnChange,  
//       
// ---------------------------------------------------------------------------
procedure TColorMemo.Loaded;
begin
  inherited Loaded;
  if Assigned(FOnChange) then FOnchange(Self);
end;

// ---------------------------------------------------------------------------
// IsChanged -      ?
// ---------------------------------------------------------------------------
function  TColorMemo.IsChanged: Boolean;
begin
  Result := (FModifyState.Text <> FLines.Text);
end;

// ---------------------------------------------------------------------------
// Change -  OnChange.      .
// ---------------------------------------------------------------------------
procedure TColorMemo.Change;
begin
  inherited Changed;

  if Assigned(FOnChange) then FOnChange(Self);
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.SetTextColor(Value: TSymbolColor);
begin
  FTextColor.Assign(Value);
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.SetSelectColor(Value: TSymbolColor);
begin
  FSelectColor.Assign(Value);
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.SetDisableColor(Value: TSymbolColor);
begin
  FDisableColor.Assign(Value);
end;

// ---------------------------------------------------------------------------
// FontChanged -    Font.OnChange
// ---------------------------------------------------------------------------
procedure TColorMemo.FontChanged(Sender: TObject);
begin
  if Assigned(FOldOnFontChange) then
    FOldOnFontChange(Sender);
  if Font.Color <> FTextColor.TextColor then
    FTextColor.TextColor := Font.Color;
end;

// ---------------------------------------------------------------------------
// TextColorChanged -   TextColor.OnColorChange
// ---------------------------------------------------------------------------
procedure TColorMemo.TextColorChanged(Sender: TObject);
begin
 if Font.Color <> FTextColor.TextColor then
    Font.Color := FTextColor.TextColor
 else
    Repaint;
end;

// ---------------------------------------------------------------------------
// SelectColorChanged -   SelectColor.OnColorChange
// ---------------------------------------------------------------------------
procedure TColorMemo.SelectColorChanged(Sender: TObject);
begin
  Repaint;
end;

// ---------------------------------------------------------------------------
// DisableColorChanged -   DisableColor.OnColorChange
// ---------------------------------------------------------------------------
procedure TColorMemo.DisableColorChanged(Sender: TObject);
begin
  Repaint;
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.SetVersion(Value: String);
begin
  ; //   !
end;

// ---------------------------------------------------------------------------
// UpdateAll -  ,   , 
// ---------------------------------------------------------------------------
procedure TColorMemo.UpdateAll(ResetScroll: Boolean);
var
  RealRow, RealCol: Longint;
begin
   PaintLock;
   try
      //    
      with FCaretState do
        CaretPosFmtToReal(RealRow, RealCol, Row, Col);
      FormatText;
      UpdateScrollRange;
      //      
      with FCaretState do
        CaretPosRealToFmt(RealRow, RealCol, Row, Col);
      if ResetScroll then
        ResetScrollPos;
      if not (csDesigning in ComponentState) then begin
        if not ResetScroll then
          MakeCaretVisible;
        UpdateCaretYPos;
        UpdateCaretXPos;
      end;
   finally
    Include(FMemoState, msPaintAll); //  
    PaintUnlock;
   end;
end;

// ---------------------------------------------------------------------------
// SetMaxLength -  .  
// ---------------------------------------------------------------------------
procedure TColorMemo.SetMaxLength(Value: Integer);
var
  FmtRow, FmtCol: Longint;
begin
  if (FMaxLength <> Value) and (FMaxLength >= 0) then
  begin
    FMaxLength := Value;
    if (FMaxLength > 0) and (FMaxLength < Length(FLines.Text)) then begin
       ClearUndoBuffer;
       Include(FMemoState, msTextChanging);
       FLines.Text := Copy(FLines.Text, 1, FMaxLength);
       Exclude(FMemoState, msTextChanging);
       if not (csDesigning in ComponentState) then begin
         FormatText;
         FmtRow := FFormatLines.Count-1;
         FmtCol := Length(GetFormatLine(FFormatLines.Count-1)) + 1;
         with FCaretState do
           CaretPosFmtToReal(Row, Col, FmtRow, FmtCol);
         MakeCaretVisible;
         UpdateCaretYPos;
         UpdateCaretXPos;
       end;
       Include(FMemoState, msPaintAll);
       Repaint;
     end;
  end;
end;

// ---------------------------------------------------------------------------
//   
// ---------------------------------------------------------------------------
procedure TColorMemo.UndoClick(Sender : TObject);
begin
  CmdUndo;
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.CutClick(Sender : TObject);
begin
  CutToClipboard;
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.CopyClick(Sender : TObject);
begin
  CopyToClipboard;
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.PasteClick(Sender : TObject);
begin
  PasteFromClipboard;
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.DeleteClick(Sender : TObject);
begin
  CmdDeleteSelected;
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.SelectAllClick(Sender : TObject);
begin
  SelectAll;
end;

// ---------------------------------------------------------------------------
// SelectAll -   
// ---------------------------------------------------------------------------
procedure TColorMemo.SelectAll;
begin
  if FLines.Count = 0 then Exit;
  PaintLock;
  try
    CmdGoFinish;
    Include(FShiftState, ssShift); //  
    PerformSelection(0, 1);
    Exclude(FShiftState, ssShift);
  finally
    PaintUnlock;
  end;

end;

// ---------------------------------------------------------------------------
//  Undo-   Command
// ---------------------------------------------------------------------------
procedure TColorMemo.SetUndoBuffer(Command: Longint);
begin
  if (FUndoBuffer.Text <> FLines.Text) and
     ((Command = CMD_UNDO) or (FUndoBuffer.LastCommand <> Command)) then begin
    FUndoBuffer.Text := FLines.Text;
    with FUndoBuffer do begin
      CaretState := FCaretState;
      SelectState := FSelectState;
      ScrollState := FScrollState;
      LastCommand := Command;
      Valid       := True;
    end;
  end;
end;

// ---------------------------------------------------------------------------
//  SetUndoCommand -    
//        
//   Undo -  (. SetUndoBuffer)
// ---------------------------------------------------------------------------
procedure TColorMemo.SetUndoCommand(Command: Longint);
begin
  FUndoBuffer.LastCommand := Command;
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.ClearUndoBuffer;
begin
  FUndoBuffer.Valid := False;
  FUndoBuffer.LastCommand := CMD_NULL;
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.InsertFirstLine;
begin
  if FLines.Count > 0 then Exit;

  Include(FMemoState, msTextChanging);
  try
    FLines.Add('');
    with FCaretState do begin
      Row := 0;
      Col := 1;
      X := LEFT_MARGIN;
      Y := TOP_MARGIN;
      FCaret.SetPos(X, Y);
    end;
    FormatText;
  finally
    Exclude(FMemoState, msTextChanging);
  end;
end;

// ---------------------------------------------------------------------------
function TColorMemo.CreateFormatLines: TFormatter;
begin
  Result := TFormatter.Create(FLines, Canvas, ClientRect,
                              LEFT_MARGIN, RIGHT_MARGIN);
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.SetWordWrap(Value: Boolean);
begin
  if Value <> FWordWrap then begin
    FWordWrap := Value;
    UpdateAll(True);
  end;
end;

// ---------------------------------------------------------------------------
// ModifyLine -       
//   Lines.OnChange: TextChanged
// ---------------------------------------------------------------------------
procedure TColorMemo.ModifyLine(Index: Longint; NewLine: String);
begin
  //    TextChanged!!!
  Include(FMemoState, msTextChanging);
  FLines[Index]:= NewLine;
  Exclude(FMemoState, msTextChanging);
end;

// ---------------------------------------------------------------------------
function TColorMemo.GetCaretHeight: Longint;
begin
  if Assigned(FCaret) then Result := FCaret.Height
  else                     Result := Canvas.TextHeight('Caret');
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.ClearCaretState;
begin
  with FCaretState do begin
    Row := 0;
    Col := 1;
    X := LEFT_MARGIN;
    Y := TOP_MARGIN;
    if FCaret <> nil then
      FCaret.SetPos(X, Y);
  end;
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.ClearScrollState;
begin
  with FScrollState do begin
    HorMinPos := 1; HorMaxPos := 1; HorPos := 1;
    VertMinPos := 1; VertMaxPos := 1; VertPos := 1;
  end;
end;

// ---------------------------------------------------------------------------
// RepaintWindow -   Repaint,    -
// PaintNeeded    
// ---------------------------------------------------------------------------
procedure TColorMemo.RepaintWindow;
begin
  if PaintNeeded then begin
    if Assigned(FCaret) then
      FCaret.Hide;
    Include(FMemoState, msRepaintWindow);
    try
      Repaint;
    finally
      if Assigned(FCaret) then
        FCaret.Show;
      Exclude(FMemoState, msRepaintWindow);
    end;
  end;
end;

// ---------------------------------------------------------------------------
// PaintNeeded -  ?
// ---------------------------------------------------------------------------
function TColorMemo.PaintNeeded: Boolean;
begin
  with FModifyState do
    Result := (Text <> FLines.Text) or
              (SelectState.RowBeg <> FSelectState.RowBeg) or
              (SelectState.ColBeg <> FSelectState.ColBeg) or
              (SelectState.RowEnd <> FSelectState.RowEnd) or
              (SelectState.ColEnd <> FSelectState.ColEnd) or
              (ScrollState.HorPos    <> FScrollState.HorPos)      or
              (ScrollState.HorMinPos <> FScrollState.HorMinPos)   or
              (ScrollState.HorMaxPos <> FScrollState.HorMaxPos)   or
              (ScrollState.VertPos    <> FScrollState.VertPos)    or
              (ScrollState.VertMinPos <> FScrollState.VertMinPos) or
              (ScrollState.VertMaxPos <> FScrollState.VertMaxPos) or
              (msPaintAll in FMemoState)
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.ClearModifyState;
begin
  if IsChanged then Change;

  with FModifyState do begin
    if HandleAllocated and (FormatLines = nil) then
      FormatLines := CreateFormatLines;

    FormatLines.Assign(FFormatLines);

    Text := FLines.Text;
    ScrollState := FScrollState;
    SelectState := FSelectState;
  end;
end;

// ---------------------------------------------------------------------------
// PaintLock -  
// ---------------------------------------------------------------------------
procedure TColorMemo.PaintLock;
begin
  Inc(FPaintLock);
end;

// ---------------------------------------------------------------------------
// PaintUnlock -  
// ---------------------------------------------------------------------------
procedure TColorMemo.PaintUnlock;
begin
  if FPaintLock > 0 then begin
    Dec(FPaintLock);
    if FPaintLock = 0 then
      RepaintWindow;
  end;
end;

// ---------------------------------------------------------------------------
// PerformSelection - ""  () 
// ---------------------------------------------------------------------------
procedure TColorMemo.PerformSelection(OldCaretRow, OldCaretCol: Longint);
var
  RowMin, RowMax: Longint;
  ColMin, ColMax: Longint;

  // ---    ?
  function SelectLeftToRight(OldCaretRow, OldCaretCol: Longint): Boolean;
  begin
    with FCaretState do
      Result := ((OldCaretRow < Row) or
                ((OldCaretRow = Row) and (OldCaretCol < Col)));
  end;

  // ---       ?
  function CaretIsLeftSelection: Boolean;
  var
    RowMin, RowMax: Longint;
    ColMin, ColMax: Longint;
  begin
    GetSelMinMaxInfo(RowMin, RowMax, ColMin, ColMax);

    with FCaretState do
      Result := ((Row < RowMin) or ((Row = RowMin) and (Col < ColMin)));
  end;

  // ---    (./.) 
  procedure SetSelLeft(R, C: Longint; var SelR, SelC: Longint);
  begin
    SelR := R;
    if Length(GetFormatLine(SelR)) = 0 then SelC := 0
    else                                    SelC := C;
  end;

  // ---    (./.) 
  procedure SetSelRight(R, C: Longint; var SelR, SelC: Longint);
  begin
    SelR := R;
    if Length(GetFormatLine(SelR)) = 0 then SelC := 0
    else                                    SelC := C - 1;
  end;

  // ---    ?
  function CaretInSelection(Row, Col: Longint): Boolean;
  var
    RowMin, RowMax: Longint;
    ColMin, ColMax: Longint;
  begin
    GetSelMinMaxInfo(RowMin, RowMax, ColMin, ColMax);
    Result := //    .  . 
              ((Row > RowMin) and (Row < RowMax))                        or
              //        
              ((Row = RowMin) and (RowMax > RowMin) and (Col >  ColMin)) or
              //        
              ((Row = RowMax) and (RowMax > RowMin) and (Col <= ColMax)) or
              //  
              ((Row = RowMin) and (Row = RowMax) and
               (Col > ColMin) and (Col <= ColMax))
  end;
  // ---
begin
  if FLines.Count = 0 then Exit; //     !
  if not InSelectMode then Exit;

  with FCaretState, FSelectState do begin
    //    
    if IsSelected then begin
       GetSelMinMaxInfo(RowMin, RowMax, ColMin, ColMax);

       //    
       if ((ColMin = ColBeg) and (Row = RowBeg) and (Col = ColBeg  )) or
          ((ColMax = ColBeg) and (Row = RowBeg) and (Col = ColBeg+1)) then begin
         ResetSelection;
         Exit;
       end;

      //   . 
      if CaretInSelection(Row, Col) then begin
        if (RowMin = RowBeg) and (ColMin = ColBeg) then
           SetSelRight(Row, Col, RowEnd, ColEnd)
        else
           SetSelLeft (Row, Col, RowEnd, ColEnd);
      end else begin
        if CaretIsLeftSelection then begin //    . 
          //        
          if (ColMin = ColBeg) and not (OldCaretCol = ColBeg) then
            Dec(ColBeg);
          SetSelLeft (Row, Col, RowEnd, ColEnd);
        end else begin                     //    . 
          //        
          if (ColMax = ColBeg) and not (OldCaretCol = ColBeg+1) then
            Inc(ColBeg);
          SetSelRight(Row, Col, RowEnd, ColEnd);
        end;
      end;
    //   
    end else begin
      if SelectLeftToRight(OldCaretRow, OldCaretCol) then begin // -
        SetSelLeft(OldCaretRow, OldCaretCol, RowBeg, ColBeg);
        SetSelRight(Row, Col, RowEnd, ColEnd);
      end else begin
        SetSelRight(OldCaretRow, OldCaretCol, RowBeg, ColBeg);
        SetSelLeft(Row, Col, RowEnd, ColEnd);
      end;
    end;
  end;

  {$IFDEF DEBUG} DebugSelection; {$ENDIF}
end;

// ---------------------------------------------------------------------------
// GetSelMinMaxInfo -    ./. /
//  
// ---------------------------------------------------------------------------
procedure TColorMemo.GetSelMinMaxInfo(var RowMin, RowMax, ColMin, ColMax: Longint);
begin
  with FSelectState do begin
    if RowBeg = RowEnd then begin
      RowMin := RowBeg;
      RowMax := RowEnd;
      ColMin := Min(ColBeg, ColEnd);
      ColMax := Max(ColBeg, ColEnd);
    end else begin
      RowMin := Min(RowBeg, RowEnd);
      RowMax := Max(RowBeg, RowEnd);
      if RowMin = RowBeg then begin
        ColMin := ColBeg;
        ColMax := ColEnd;
      end else begin
        ColMin := ColEnd;
        ColMax := ColBeg;
      end;
    end;
  end;
end;

// ---------------------------------------------------------------------------
// IsSelectedChar -    ?
// ---------------------------------------------------------------------------
function TColorMemo.IsSelectedChar(Row, Col: Longint): Boolean;
var
  RowMin, RowMax: Longint;
  ColMin, ColMax: Longint;
begin
  GetSelMinMaxInfo(RowMin, RowMax, ColMin, ColMax);
  Result := //    .  . 
            ((Row > RowMin) and (Row < RowMax))                        or
            //        
            ((Row = RowMin) and (RowMax > RowMin) and (Col >= ColMin)) or
            //        
            ((Row = RowMax) and (RowMax > RowMin) and (Col <= ColMax)) or
            //  
            ((Row = RowMin) and (Row = RowMax) and
             (Col >= ColMin) and (Col <= ColMax))
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.CheckSelectMode;
begin
  if not InSelectMode then
     ResetSelection;
end;

// ---------------------------------------------------------------------------
// InSelectMode -      ?
// ---------------------------------------------------------------------------
function  TColorMemo.InSelectMode: Boolean;
begin
  Result := (ssShift in FShiftState) or (msMouseSelection in FMemoState);
end;

// ---------------------------------------------------------------------------
// IsSelected -   ?
// ---------------------------------------------------------------------------
function TColorMemo.IsSelected: Boolean;
begin
  //   Col = 0,    
  with FSelectState do
    Result := (RowBeg >= 0) and (RowEnd >= 0) and
              (ColBeg > -1) and (ColEnd > -1);
end;

// ---------------------------------------------------------------------------
// IsAllSelected -   ?
// ---------------------------------------------------------------------------
function TColorMemo.IsAllSelected: Boolean;
var
  RowMin, RowMax, ColMin, ColMax: Longint;
begin
  GetSelMinMaxInfo(RowMin, RowMax, ColMin, ColMax);
  Result := (RowMin = 0) and (ColMin = 1) and
            (RowMax = FFormatLines.Count-1) and
            (ColMax = Length(GetFormatLine(FFormatLines.Count-1)));
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.ResetSelection;
begin
  // Col = 0   
  with FSelectState do begin
    RowBeg := -1; ColBeg := -1;
    RowEnd := -1; ColEnd := -1;
  end;
end;


// ***************************************************************************
//  
// ***************************************************************************
{$IFNDEF DELPHI_400}
function TColorMemo.StrToOem(S: String): String;
begin
  Result := S;
end;
{$ENDIF}

procedure TColorMemo.DebugLn(S: String);
begin
  {$IFDEF DEBUG}
  WriteLn(StrToOem(S));
  {$ENDIF}
end;


{$IFDEF DEBUG}
procedure TColorMemo.DebugTextStatistic;
begin
  {$IFDEF DEBUG_TEXT_STATISTIC}
  WriteLn('');
  WriteLn('Debug: Text Statistic:');
  WriteLn('----------------------');
  with FTextStatistic do begin
    WriteLn('  TotalWidth:    ' + IntToStr(TotalWidth));
    WriteLn('  AvgTailCharW:  ' + FloatToStr(AvgTailCharW));
    WriteLn('  TailCharCount: ' + IntToStr(TailCharCount));
    WriteLn('  VisibleRows:   ' + IntToStr(VisibleRows));
    WriteLn('  VertPageSize:  ' + IntToStr(VertPageSize));
    WriteLn('  VisibleChars:  ' + IntToStr(VisibleChars));
  end;
  {$ENDIF}
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.DebugCaretPos;
begin
  {$IFDEF DEBUG_CARET}
  WriteLn('');
  WriteLn('Debug: Caret Position:');
  WriteLn('----------------------');
  with FCaretState do begin
    WriteLn('  Row: ' + IntToStr(Row));
    WriteLn('  Col: ' + IntToStr(Col));
  end;
  {$ENDIF}
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.DebugSelection;
var
  RowMin, RowMax, ColMin, ColMax: Longint;
begin
  {$IFDEF DEBUG_SELECTION}
  GetSelMinMaxInfo(RowMin, RowMax, ColMin, ColMax);
  WriteLn('');
  WriteLn('Debug: Selection:');
  WriteLn('-----------------');
  WriteLn('  RowMin: ' + IntToStr(RowMin) + ', RowMax: ' + IntToStr(RowMax));
  WriteLn('  ColMin: ' + IntToStr(ColMin) + ', ColMax: ' + IntToStr(ColMax));
 {$ENDIF}
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.DebugSelectionReal;
var
  FmtRowMin, FmtRowMax, FmtColMin, FmtColMax: Longint;
  RowMin, RowMax, ColMin, ColMax: Longint;
begin
  {$IFDEF DEBUG_SELECTION}
  GetSelMinMaxInfo(FmtRowMin, FmtRowMax, FmtColMin, FmtColMax);
  CaretPosFmtToReal(RowMin, ColMin, FmtRowMin, FmtColMin);
  CaretPosFmtToReal(RowMax, ColMax, FmtRowMax, FmtColMax);

  WriteLn('');
  WriteLn('Debug: Real Selection:');
  WriteLn('----------------------');
  WriteLn('RowMin: ' + IntToStr(RowMin) + ', RowMax: ' + IntToStr(RowMax));
  WriteLn('ColMin: ' + IntToStr(ColMin) + ', ColMax: ' + IntToStr(ColMax));
  {$ENDIF}
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.DebugFormatLines;
var
  I: Longint;
begin
  WriteLn('');
  WriteLn('Debug: Format Lines:');
  WriteLn('--------------------');
  if not Assigned(FFormatLines) then Exit;

  for I := 0 to FFormatLines.Count-1 do
    DebugLn( '  ' + GetFormatLine(I));
  WriteLn('--------------------');
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.DebugModifyFormatLines;
var
  I: Longint;
  P: PFormatLine;
begin
  WriteLn('');
  WriteLn('Debug: Modify Format Lines:');
  WriteLn('---------------------------');
  if not Assigned(FModifyState.FormatLines) then Exit;

  for I := 0 to FModifyState.FormatLines.Count-1 do begin
    P := PFormatLine(FModifyState.FormatLines.Items[I]);
    DebugLn( '  ' + P^.Line);
  end;
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.DebugScrollRange;
var
  ScrollInfo: TScrollInfo;
begin
  WriteLn('');
  WriteLn('Debug: ScrollRange:');
  WriteLn('-------------------');
  with FScrollState do begin
    WriteLn('  VertMinPos: ' + IntToStr(VertMinPos));
    WriteLn('  VertPos:    ' + IntToStr(VertPos));
    WriteLn('  VertMaxPos: ' + IntToStr(VertMaxPos));
    WriteLn('  HorMinPos: ' + IntToStr(HorMinPos));
    WriteLn('  HorPos:    ' + IntToStr(HorPos));
    WriteLn('  HorMaxPos: ' + IntToStr(HorMaxPos));

    with ScrollInfo do begin
      fMask := SIF_PAGE or SIF_RANGE or SIF_POS;
      GetScrollInfo(Handle, SB_VERT, ScrollInfo);
      WriteLn('nMin : ' + IntToStr(nMin));
      WriteLn('nMax : ' + IntToStr(nMax));
      WriteLn('nPage: ' + IntToStr(nPage));
      WriteLn('nPos : ' + IntToStr(nPos));
    end;
  end;
end;

{$ENDIF}

// ***************************************************************************
//  
// ***************************************************************************

// ---------------------------------------------------------------------------
// CaretPosFmtToReal -     
//    
// ---------------------------------------------------------------------------
procedure TColorMemo.CaretPosFmtToReal(var RealRow, RealCol: Longint;
                                       FmtRow, FmtCol: Longint);
var
  P: PFormatLine;
begin
  if csDesigning in ComponentState then
    Exit;
  // FmtCol   0!  Select-
  if (FmtRow < 0) or (FmtCol < 0) then
    Exit;
  if (FFormatLines = nil) or (FmtRow > FFormatLines.Count-1) then
    Exit;

  P := PFormatLine(FFormatLines.Items[FmtRow]);
  RealRow := P^.OrigLineIndex;
  if P^.BegPos > 0 then RealCol := P^.BegPos - 1 + FmtCol
  else                  RealCol := 1; //   !!!
end;

// ---------------------------------------------------------------------------
// CaretPosFmtToReal -     
//    
// ---------------------------------------------------------------------------
procedure TColorMemo.CaretPosRealToFmt(RealRow, RealCol: Longint;
                                       var FmtRow, FmtCol: Longint);
var
  P: PFormatLine;
  I: Longint;
begin
  if csDesigning in ComponentState then
    Exit;
  // RealCol   0!  Select-
  if (RealRow < 0) or (RealCol < 0) then
    Exit;
  if (FFormatLines = nil) or (RealRow > Lines.Count-1) then
    Exit;

  with FFormatLines do
    for I := 0 to Count-1 do begin
      P := PFormatLine(Items[I]);
      if (P^.OrigLineIndex = RealRow) and
         (P^.BegPos <= RealCol) and (RealCol <= P^.EndPos+1) then
      begin
        FmtRow := I;
        if P^.BegPos > 0 then FmtCol := RealCol - P^.BegPos + 1
        else                  FmtCol := 1; //   !!!
      end;
    end;
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.CopyToClipboard;
begin
  if HandleAllocated then
    Perform(WM_COPY, 0, 0);
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.CutToClipboard;
begin
  if HandleAllocated then
    Perform(WM_CUT, 0, 0);
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.PasteFromClipboard;
begin
  if HandleAllocated then
    Perform(WM_PASTE, 0, 0);
end;

// ---------------------------------------------------------------------------
// GetSelText -    
// ---------------------------------------------------------------------------
function TColorMemo.GetSelText: String;
var
  RowMin, RowMax, ColMin, ColMax: Longint;
  FmtRowMin, FmtRowMax, FmtColMin, FmtColMax: Longint;
  I, Len: Longint;
  EndStr: String;
begin
  Result := '';
  if not IsSelected then Exit;

  GetSelMinMaxInfo(FmtRowMin, FmtRowMax, FmtColMin, FmtColMax);

  //    -   ,    ,
  //        
  CaretPosFmtToReal(RowMin, ColMin, FmtRowMin, FmtColMin);
  CaretPosFmtToReal(RowMax, ColMax, FmtRowMax, FmtColMax);

  for I := 0 to FLines.Count-1 do begin
    if (I > RowMin) and (I < RowMax) then
      Result := Result + FLines[I] + CRLF
    else if I = RowMin then begin
      EndStr := '';
      if RowMax = RowMin then
        Len := ColMax - ColMin + 1
      else begin
        Len := MAX_STR_LEN;
        EndStr := CRLF;
      end;
      Result := Result + Copy(FLines[I], ColMin, Len) + EndStr;
    end else if I = RowMax then begin
      if RowMax > RowMin then
        Result := Result + Copy(FLines[I], 1, ColMax);
    end;
  end;
{
  WriteLn('FmtRowMin: ' + IntToStr(FmtRowMin) + ', FmtColMin: ' + IntToSTr(FmtColMin));
  WriteLn('FmtRowMax: ' + IntToStr(FmtRowMax) + ', FmtColMax: ' + IntToSTr(FmtColMax));
  WriteLn('RowMin: ' + IntToStr(RowMin) + ', ColMin: ' + IntToStr(ColMin));
  WriteLn('RowMax: ' + IntToStr(RowMax) + ', ColMax: ' + IntToStr(ColMax));
  WriteLn(StrToOEM(Result));
}  
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.CmdCopy;
begin
  with Clipboard do begin
    Open;
    try
      Clear;
      SetTextBuf(PChar(GetSelText));
    finally
      Close;
    end;
  end;
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.CmdCut;
begin
  if FReadOnly then Exit;

  SetUndoBuffer(CMD_CUT);
  CmdCopy;
  CmdDeleteSelected;
end;

// ---------------------------------------------------------------------------
//      1.20
const
     CLPBUF_LEN = 30*1024;

var
  CharBuf: array [0..CLPBUF_LEN-1] of Char;


// ---------------------------------------------------------------------------
// CmdPaste -    Cliopboard
//               1.20,    
//             !
// ---------------------------------------------------------------------------
procedure TColorMemo.CmdPaste;
var
  BufLen, I, I0: Longint;
  OldCaretState: TCaretState;
  CaretRow, CaretCol: Longint;
  S, StrLeft, StrRight: String;
begin
  if FReadOnly then Exit;

  if FLines.Count = 0 then
    InsertFirstLine;


  FCaret.Hide;
  PaintLock;
  Include(FMemoState, msTextChanging);
  try
    if IsSelected then begin
      CmdDeleteSelected;
      SetUndoCommand(CMD_CHAR);
    end;
    SetUndoBuffer(CMD_CHAR);

    with Clipboard do begin
      Open;
      try
        BufLen := GetTextBuf(CharBuf, CLPBUF_LEN-2);
        OldCaretState := FCaretState; //  - PerformSelection

        with FCaretState do
          CaretPosFmtToReal(CaretRow, CaretCol, Row, Col);

        S := Lines[CaretRow];
        StrLeft  := Copy(S, 1, CaretCol-1);
        StrRight := Copy(S, CaretCol, 255);

        I := 0;
        S := '';
        while (I < BufLen) and  (CharBuf[I] <> #13) do
          Inc(I);

        SetString(S, CharBuf, I);
        StrLeft := StrLeft + S;
        Lines[CaretRow] := StrLeft;

        Inc(I);
        if (I  < BufLen) and (CharBuf[I] = #10) then
           Inc(I);

        while I < BufLen do begin
          I0 := I;
          while (I < BufLen) and  (CharBuf[I] <> #13) do
            Inc(I);

          SetString(S, PChar(@CharBuf[I0]), I-I0);
          Lines.Insert(CaretRow+1, S);

          Inc(I);
          if (I  < BufLen) and (CharBuf[I] = #10) then
            Inc(I);

          Inc(CaretRow);
        end;

        CaretCol := Length(Lines[CaretRow]) + 1;
        Lines[CaretRow] := Lines[CaretRow] + StrRight;

        FormatText;

        with FCaretState do
          CaretPosRealToFmt(CaretRow, CaretCol, Row, Col);

        UpdateScrollRange;
        UpdateCaretYPos;
        UpdateCaretXPos;

        if not CaretIsVertVisible then
          //ScrollVert(FCaretState.Row + 1);
          ScrollVert(FCaretState.Row - FTextStatistic.VisibleRows + 2);

        //     
        Include(FShiftState, ssShift); //   
        with OldCaretState do
          PerformSelection(Row, Col);
      finally
        Close;
      end;
    end;


  finally
    Exclude(FMemoState, msTextChanging);
    PaintUnlock;
    FCaret.Show;
  end;
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.CmdUndo;
var
  Buffer: TUndoRecord;
begin
  if not FUndoBuffer.Valid then Exit;

  Buffer := FUndoBuffer;

  SetUndoBuffer(CMD_UNDO);
  PaintLock;
  try
    Include(FMemoState, msTextChanging);
    FLines.Text := Buffer.Text;
    Exclude(FMemoState, msTextChanging);
    with Buffer do begin
      FCaretState := CaretState;
      FSelectState := SelectState;
      FScrollState := ScrollState;
    end;

    FormatText;
    UpdateScrollRange;
    UpdateCaretYPos;
    UpdateCaretXPos;
  finally
    Include(FMemoState, msPaintAll);
    PaintUnlock;
  end;
end;


// ---------------------------------------------------------------------------
procedure TColorMemo.CmdDeleteSelected;
var
  FmtRowMin, FmtRowMax, FmtColMin, FmtColMax: Longint;
  RowMin, ColMin, RowMax, ColMax: Longint;
  I: Longint;
  CaretRow, CaretCol: Longint;
  Line1, Line2: String;
  NewLines: TStringList;
  Cnt: Longint;
  SaveIndex: Longint;
begin
  if not IsSelected or FReadOnly then Exit;

  SetUndoBuffer(CMD_CUT);

  GetSelMinMaxInfo(FmtRowMin, FmtRowMax, FmtColMin, FmtColMax);
  //    -   ,    ,
  //        
  CaretPosFmtToReal(RowMin, ColMin, FmtRowMin, FmtColMin);
  CaretPosFmtToReal(RowMax, ColMax, FmtRowMax, FmtColMax);

  {$IFDEF DEBUG} DebugSelectionReal; {$ENDIF}

  NewLines := TStringList.Create;
  Include(FMemoState, msTextChanging);
  PaintLock;
  try
    SaveIndex := -1;
    Line1 := '';
    Line2 := '';
    CaretRow := 0;
    CaretCol := 1;
    for I := 0 to FLines.Count-1 do
      if (I < RowMin) or (I > RowMax) then
        NewLines.Add(FLines[I])
      else if I = RowMin then begin
        if RowMin = RowMax then Cnt := ColMax - ColMin + 1
        else                    Cnt := MAX_STR_LEN;
        Line1 := FLines[I];
        Delete(Line1, ColMin, Cnt);
        SaveIndex := I;
        CaretRow := I;
        CaretCol := ColMin;
      end else if I = RowMax then begin
        Line2 := Copy(FLines[I], ColMax+1, MAX_STR_LEN);
      end;

    if SaveIndex >= 0 then begin
      Line1 := Line1 + Line2;
      NewLines.Insert(SaveIndex, Line1);
    end;

    FLines.Free;
    FLines := NewLines;
    TStringList(FLines).OnChange := TextChanged;

    FormatText;

    with FCaretState do
      CaretPosRealToFmt(CaretRow, CaretCol, Row, Col);

    UpdateScrollRange;
    UpdateCaretYPos;
    UpdateCaretXPos;
    ResetSelection;

    if not CaretIsVertVisible then
      ScrollVert(FCaretState.Row + 1);

  finally
    Exclude(FMemoState, msTextChanging);
    Include(FMemoState, msPaintAll);
    PaintUnlock;
  end;

  {$IFDEF DEBUG} DebugCaretPos; {$ENDIF}
end;

// ---------------------------------------------------------------------------
// CmdDeleteChar -  
// Dir < 0 -   , Dir > 0 -   
// ---------------------------------------------------------------------------
procedure TColorMemo.CmdDeleteChar(Dir: Longint);
var
  Line: String;
  RealRow, RealCol: Longint;

  procedure DoFormat;
  begin
    FormatText;
    UpdateScrollRange;
    with FCaretState do
      CaretPosRealToFmt(RealRow, RealCol, Row, Col);
    UpdateCaretYPos;
    UpdateCaretXPos;
  end;
begin
  if FReadOnly then Exit;
  if FLines.Count = 0 then Exit;

  with FCaretState do
    CaretPosFmtToReal(RealRow, RealCol, Row, Col);
  Line := FLines[RealRow];

  PaintLock;
  try
    with FCaretState do
      if Dir < 0 then begin
        if (RealCol > 1) or (RealRow > 0) then
          SetUndoBuffer(CMD_CUT);
        if RealCol > 1 then begin
          Delete(Line, RealCol-1, 1);
          ModifyLine(RealRow, Line);
          Dec(RealCol);
          DoFormat;
        end else if RealRow > 0 then begin
          CmdCombineLines(RealRow-1, RealRow);
        end;
      end else begin
        if (RealCol <= Length(Line)) or (RealRow < FLines.Count-1) then
          SetUndoBuffer(CMD_CUT);
        if RealCol <= Length(Line) then begin
          Delete(Line, RealCol, 1);
          ModifyLine(RealRow, Line);
          DoFormat;
        end else if RealRow < FLines.Count-1 then begin
          CmdCombineLines(RealRow, RealRow+1);
        end;
      end;

      if not CaretIsVertVisible then
        if FCaretState.Row < GetFirstRow then
          ScrollVert(FScrollState.VertPos - 1)
        else
          ScrollVert(FScrollState.VertPos + 1);
  finally
    PaintUnlock;
  end;
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.CmdNewLine;
var
  Line, Rest: String;
  RealRow, RealCol: Longint;
begin
  if FReadOnly then Exit;

  if FLines.Count = 0 then
    InsertFirstLine;

  SetUndoBuffer(CMD_CHAR);

  PaintLock;
  Include(FMemoState, msTextChanging);
  try
    with FCaretState do begin
      with FCaretState do
        CaretPosFmtToReal(RealRow, RealCol, Row, Col);
      Line := FLines[RealRow];
      Rest := Copy(Line, RealCol, MAX_STR_LEN);
      FLines.Insert(RealRow+1, Rest);
      Delete(Line, RealCol, MAX_STR_LEN);
      FLines[RealRow] := Line;
      RealCol := 1;
      Inc(RealRow);

      FormatText;
      with FCaretState do
        CaretPosRealToFmt(RealRow, RealCol, Row, Col);
      UpdateScrollRange;
      UpdateCaretYPos;
      UpdateCaretXPos;

      if not CaretIsVertVisible then
         ScrollVert(FScrollState.VertPos + 1);
      MakeRowBegVisible;
    end;
  finally
    Exclude(FMemoState, msTextChanging);
    PaintUnlock;
  end;
end;

// ---------------------------------------------------------------------------
// CmdCombineLines -   
// ---------------------------------------------------------------------------
procedure TColorMemo.CmdCombineLines(StartRow, EndRow: Longint);
var
  Line: String;
  RealRow, RealCol: Longint;
  Len: Longint;
begin
  Include(FMemoState, msTextChanging);
  try
    Line := FLines[StartRow];
    Len := Length(Line);
    Line := Line + FLines[EndRow];
    FLines[StartRow] := Line;
    FLines.Delete(EndRow);
    RealRow := StartRow;
    RealCol := Len + 1;

    FormatText;
    with FCaretState do
      CaretPosRealToFmt(RealRow, RealCol, Row, Col);
    UpdateScrollRange;
    UpdateCaretYPos;
    UpdateCaretXPos;
  finally
    Exclude(FMemoState, msTextChanging);
  end;
end;

// ---------------------------------------------------------------------------
// CmdChar -  .
// ---------------------------------------------------------------------------
procedure TColorMemo.CmdChar(CharCode: Word);
var
  RealRow, RealCol: Longint;
  RealLine: String;
  InsChr: Char;
  InsStr: String;
begin
  if FReadOnly then Exit;

  if CharCode in [0, VK_BACK, VK_ESCAPE, VK_RETURN] then Exit;

  if not (Char(CharCode) in [#32..#255]) then begin
    MessageBeep(Cardinal(-1));
    Exit;
  end;

  FShiftState := [];

  if (FMaxLength > 0) and (Length(FLines.Text) >= FMaxLength + 2) then
    Exit;


  //     !
  if FLines.Count = 0 then
    InsertFirstLine;

  PaintLock;
  try
    if IsSelected then begin
      CmdDeleteSelected;
      SetUndoCommand(CMD_CHAR);
    end;

    SetUndoBuffer(CMD_CHAR);

    with FCaretState do
      CaretPosFmtToReal(RealRow, RealCol, Row, Col);

    InsChr := Char(CharCode);
    SetString(InsStr, PChar(@InsChr), 1);

    RealLine := FLines[RealRow];
    Insert(InsStr, RealLine, RealCol);

    ModifyLine(RealRow, RealLine);
    FormatText;
    UpdateScrollRange;

    with FCaretState do
      CaretPosRealToFmt(RealRow, RealCol, Row, Col);

    if not CaretIsVertVisible then
       ScrollVert(FScrollState.VertPos + 1);

    UpdateCaretYPos;
    UpdateCaretXPos;
    CmdGoRight;
  finally
    PaintUnlock;
  end;
end;

// ***************************************************************************
//   
// ***************************************************************************

// ---------------------------------------------------------------------------
// CmdSelectWord -   (,    )
// ---------------------------------------------------------------------------
procedure TColorMemo.CmdSelectWord;
const
  EMPTY_CHAR = [#32, #9]; // , 
var
  WordBegPos, WordEndPos: Longint;
  Len: Longint;
  Line: String;
begin
  if FLines.Count = 0 then Exit;

  Line := GetFormatLine(FCaretState.Row);
  Len := Length(Line);
  if Len = 0 then Exit;

  WordBegPos := FCaretState.Col - 1;
  if WordBegPos = 0 then Inc(WordBegPos);

  while WordBegPos > 1 do begin
    if not (Line[WordBegPos] in EMPTY_CHAR) then
      break;
    Dec(WordBegPos);
  end;
  while WordBegPos > 1 do begin
    if Line[WordBegPos-1] in EMPTY_CHAR then
      break;
    Dec(WordBegPos);
  end;

  WordEndPos := FCaretState.Col - 1;
  if WordEndPos = 0 then Inc(WordEndPos);
  while WordEndPos < Len do begin
    if Line[WordEndPos] in EMPTY_CHAR then
      break;
    Inc(WordEndPos);
  end;
  while WordBegPos < Len do begin
    if not (Line[WordEndPos+1] in EMPTY_CHAR) then
      break;
    Inc(WordEndPos);
  end;

  with FSelectState do begin
    FCaretState.Col := WordEndPos + 1;
    RowBeg := FCaretState.Row;
    RowEnd := RowBeg;
    ColBeg := WordBegPos;
    ColEnd := WordEndPos;
  end;

  RepaintWindow;
  UpdateCaretXPos;
end;

// ---------------------------------------------------------------------------
// CmdGoStart -    
// ---------------------------------------------------------------------------
procedure TColorMemo.CmdGoStart;
var
  OldCaretRow, OldCaretCol: Longint;
begin
  if Lines.Count = 0 then Exit;
  CheckSelectMode;

  PaintLock;
  try
    with FScrollState, FCaretState do begin
      OldCaretRow := Row;
      OldCaretCol := Col;

      if VertPos > VertMinPos then
        ScrollVert(VertMinPos);   //   
      Row := 0;
      Col := 1;
      MakeRowBegVisible;
      UpdateCaretYPos;
      UpdateCaretXPos;

      if InSelectMode then  //  
        PerformSelection(OldCaretRow, OldCaretCol);
    end;
  finally
     PaintUnlock;
  end;
end;

// ---------------------------------------------------------------------------
// CmdGoFinish -    .
// ---------------------------------------------------------------------------
procedure TColorMemo.CmdGoFinish;
var
  OldCaretRow, OldCaretCol: Longint;
begin
  if Lines.Count = 0 then Exit;
  CheckSelectMode;

  PaintLock;
  try
    with FScrollState, FCaretState do begin
      OldCaretRow := Row;
      OldCaretCol := Col;

      if VertPos < VertMaxPos then
        ScrollVert(VertMaxPos);  //   
      Row := FFormatLines.Count-1;
      Col := Length(GetFormatLine(Row)) + 1;
      MakeRowEndVisible;
      UpdateCaretYPos;
      UpdateCaretXPos;
    end;

    if InSelectMode then  //  
      PerformSelection(OldCaretRow, OldCaretCol);
  finally
     PaintUnlock;
  end;
end;

// ---------------------------------------------------------------------------
// CmdGoHome -    .
// ---------------------------------------------------------------------------
procedure TColorMemo.CmdGoHome;
var
  OldCaretRow, OldCaretCol: Longint;
begin
  if Lines.Count = 0 then Exit;
  CheckSelectMode;

  PaintLock;
  try
    with FCaretState, FScrollState do begin
      OldCaretRow := Row;
      OldCaretCol := Col;

      MakeRowBegVisible;
      Col := 1;
      X := LEFT_MARGIN;
      FCaret.SetPos(X, Y);
    end;

    if InSelectMode then  //  
      PerformSelection(OldCaretRow, OldCaretCol);
  finally
    PaintUnlock;
  end;
end;

// ---------------------------------------------------------------------------
// CmdGoEnd -    .
// ---------------------------------------------------------------------------
procedure TColorMemo.CmdGoEnd;
var
  OldCaretRow, OldCaretCol: Longint;
begin
  if Lines.Count = 0 then Exit;
  CheckSelectMode;

  PaintLock;
  try
    with FCaretState, FScrollState do begin
      OldCaretRow := Row;
      OldCaretCol := Col;

      MakeRowEndVisible;
      Col := GetRowEndPos(Row, X);
      FCaret.SetPos(X, Y);
    end;

    if InSelectMode then  //  
      PerformSelection(OldCaretRow, OldCaretCol);
  finally
    PaintUnlock;
  end;
end;

// ---------------------------------------------------------------------------
// CmdGoLeft -    
// ---------------------------------------------------------------------------
procedure TColorMemo.CmdGoLeft;
var
  Line: String;
  OldCaretRow, OldCaretCol: Longint;
begin
  if Lines.Count = 0 then Exit;
  CheckSelectMode;

  PaintLock;
  try
    with FCaretState, FScrollState do begin
      OldCaretRow := Row;
      OldCaretCol := Col;

      Line := GetFormatLine(Row);
      if Col = 1 then begin //     
        if Row = 0 then Exit;
        CmdGoUp;
        MakeRowEndVisible;
        Col := GetRowEndPos(Row, X);
        FCaret.SetPos(X, Y);
      end else begin
        Dec(Col);
        X := X - Canvas.TextWidth(Line[Col]);
        if X < LEFT_MARGIN then begin
          ScrollHor(HorPos - HOR_SCROLL_CHUNK);
          UpdateCaretXPos;
        end else
          FCaret.SetPos(X, Y);
      end;
    end;

    if InSelectMode then  //  
      PerformSelection(OldCaretRow, OldCaretCol);
  finally
    PaintUnlock;
  end;
end;

// ---------------------------------------------------------------------------
// CmdGoRight -    
// ---------------------------------------------------------------------------
procedure TColorMemo.CmdGoRight;
var
  Line: String;
  OldCaretRow, OldCaretCol: Longint;
begin
  if Lines.Count = 0 then Exit;
  CheckSelectMode;

  PaintLock;
  try
    with FCaretState, FScrollState do begin
      OldCaretRow := Row;
      OldCaretCol := Col;

      Line := GetFormatLine(Row);
      if Col = Length(Line) + 1 then begin //     
        if Row = FFormatLines.Count-1 then Exit;
        CmdGoDown;
        MakeRowBegVisible;
        Col := 1;
        X := LEFT_MARGIN;
        FCaret.SetPos(X, Y);
      end else begin
        X := X + Canvas.TextWidth(Line[Col]);
        Inc(Col);
        if X > (ClientRect.Right - RIGHT_MARGIN) then begin
          ScrollHor(HorPos + HOR_SCROLL_CHUNK);
          UpdateCaretXPos;
        end else
          FCaret.SetPos(X, Y);
       end;
    end;

    if InSelectMode then //  
      PerformSelection(OldCaretRow, OldCaretCol);
  finally
    PaintUnlock;
  end;
end;

// ---------------------------------------------------------------------------
// CmdGoUp -    
// ---------------------------------------------------------------------------
procedure TColorMemo.CmdGoUp;
var
  FirstRow: Longint;
  OldCaretRow, OldCaretCol: Longint;
begin
  if Lines.Count = 0 then Exit;
  CheckSelectMode;

  PaintLock;
  try
    MakeCaretVisible;
    with FCaretState, FScrollState do begin
      OldCaretRow := Row;
      OldCaretCol := Col;

      FirstRow := GetFirstRow;
      if (FirstRow = 0) and (Row = FirstRow) then
        Exit;
      if Row = FirstRow then
        ScrollVert(VertPos-1);
      PositionCaretXY(X, Y - GetCaretHeight + 1);
      // MakeCaretVisible    1.20  
      //       
      //    ->, <-
    end;

    if InSelectMode then  //  
      PerformSelection(OldCaretRow, OldCaretCol);
  finally
    PaintUnlock;
  end;

end;

// ---------------------------------------------------------------------------
// CmdGoPgUp -    
//                1.20
// ---------------------------------------------------------------------------
procedure TColorMemo.CmdGoPgUp;
var
  I: Longint;
  Xpos,YPos: Longint;
  OldCaretRow, OldCaretCol: Longint;
begin
  if Lines.Count = 0 then Exit;
  if FScrollState.VertPos = FScrollState.VertMinPos then
    Exit;
  CheckSelectMode;

  with FCaretState do begin
    XPos := X;
    YPos := Y;
    OldCaretRow := Row;
    OldCaretCol := Col;
  end;

  with FScrollState do
    ScrollVert(VertPos - FTextStatistic.VisibleRows + 1);

  PositionCaretXY(XPos, YPos);

  FCaret.Show;

  if InSelectMode then begin
    PerformSelection(OldCaretRow, OldCaretCol);
    RepaintWindow;
  end;

end;

// ---------------------------------------------------------------------------
// CmdGoDown -    
// ---------------------------------------------------------------------------
procedure TColorMemo.CmdGoDown;
var
  LastRow: Longint;
  OldCaretRow, OldCaretCol: Longint;
begin
  if Lines.Count = 0 then Exit;
  CheckSelectMode;

  PaintLock;
  try
    MakeCaretVisible;
    with FCaretState, FScrollState, FTextStatistic do begin
      OldCaretRow := Row;
      OldCaretCol := Col;

      LastRow := GetFirstRow + VisibleRows - 1;
      if (LastRow = FFormatLines.Count-1) and (Row = LastRow) then Exit;
      if Row = LastRow then
        ScrollVert(VertPos+1);
      PositionCaretXY(X, Y + GetCaretHeight + 1);
      // MakeCaretVisible    1.20  
      //       
      //    ->, <-
    end;

    if InSelectMode then  //  
      PerformSelection(OldCaretRow, OldCaretCol);
  finally
    PaintUnlock;
  end;
end;

// ---------------------------------------------------------------------------
// CmdGoPgDown -    
//                  1.20
// ---------------------------------------------------------------------------
procedure TColorMemo.CmdGoPgDown;
var
  I: Longint;
  XPos,YPos: Longint;
  OldCaretRow, OldCaretCol: Longint;
begin
  if Lines.Count = 0 then Exit;
  if FScrollState.VertPos = FScrollState.VertMaxPos then
    Exit;
  CheckSelectMode;

  with FCaretState do begin
    XPos := X;
    YPos := Y;
    OldCaretRow := Row;
    OldCaretCol := Col;
  end;

  with FScrollState do
    ScrollVert(VertPos + FTextStatistic.VisibleRows - 1);

  PositionCaretXY(XPos, YPos);

  FCaret.Show;

  if InSelectMode then begin
    PerformSelection(OldCaretRow, OldCaretCol);
    RepaintWindow;
  end;
end;

// ---------------------------------------------------------------------------
// MakeRowBegVisible -     , 
//    
// ---------------------------------------------------------------------------
procedure TColorMemo.MakeRowBegVisible;
begin
  if FLines.Count = 0 then Exit;

  with FScrollState do
    if HorPos > HorMinPos then
      ScrollHor(HorMinPos);
end;

// ---------------------------------------------------------------------------
// MakeRowEndVisible -     , 
//    
// ---------------------------------------------------------------------------
procedure TColorMemo.MakeRowEndVisible;
var
  I: Longint;
  OldHorPos, NewHorPos: Longint;
  Offset, TextW: Longint;
begin
  if FLines.Count = 0 then Exit;

  with FScrollState do begin
    OldHorPos := HorPos;
    TextW := Canvas.TextWidth(GetFormatLine(FCaretState.Row));
    for I := HorMinPos to HorMaxPos do begin
      HorPos := I;
      Offset := GetHorOffset;
      if (LEFT_MARGIN + TextW + RIGHT_MARGIN <= Offset + ClientRect.Right) then
        break;
    end;
    NewHorPos := HorPos;
    HorPos := OldHorPos;

    if NewHorPos > HorPos then
      ScrollHor(NewHorPos);
  end;
end;

// ---------------------------------------------------------------------------
// CaretIsHorVisible -     ?
// ---------------------------------------------------------------------------
function  TColorMemo.CaretIsHorVisible: Boolean;
begin
  with FCaretState do
    Result := (X >= LEFT_MARGIN) and (X <= ClientRect.Right - RIGHT_MARGIN);
end;

// ---------------------------------------------------------------------------
// MakeCaretVisible -     , 
//   .
// ---------------------------------------------------------------------------
procedure TColorMemo.MakeCaretVisible;
var
 Off, HP: Longint;
begin
  if FLines.Count = 0 then Exit;

  with FCaretState, FScrollState, FTextStatistic do begin
    if not CaretIsVertVisible then
      ScrollVert(VertMinPos + FCaretState.Row);
    if not CaretIsHorVisible then begin
        if X < LEFT_MARGIN then begin
          Off := LEFT_MARGIN - X;
          HP := HorMinPos;
          if AvgTailCharW > 0.00 then
            HP := HorPos - Trunc(Off/AvgTailCharW + 0.5) - HOR_SCROLL_CHUNK;
        end else begin
          Off := X - ClientRect.Right + RIGHT_MARGIN;
          HP := HorMaxPos;
          if AvgTailCharW > 0.00 then
            HP := HorPos + Trunc(Off/AvgTailCharW + 0.5) + HOR_SCROLL_CHUNK;
        end;
        ScrollHor(HP);
    end;
  end;

end;

// ---------------------------------------------------------------------------
// GetFirstRow -     .
// ---------------------------------------------------------------------------
function TColorMemo.GetFirstRow: Longint;
begin
  Result := -1;
  if FLines.Count = 0 then Exit;

  with FScrollState do
    Result := VertPos - VertMinPos;
end;

// ---------------------------------------------------------------------------
// ScrollVert -   
// ---------------------------------------------------------------------------
procedure TColorMemo.ScrollVert(Pos: Smallint);
begin
  with FScrollState do begin
    if Pos < VertMinPos then Pos := VertMinPos;
    if Pos > VertMaxPos then Pos := VertMaxPos;
  end;

  Perform(WM_VSCROLL,  (Pos shl 16) or SB_THUMBPOSITION, 0);
end;

// ---------------------------------------------------------------------------
// ScrollHor -   
// ---------------------------------------------------------------------------
procedure TColorMemo.ScrollHor(Pos: Smallint);
begin
  with FScrollState do begin
    if Pos < HorMinPos then Pos := HorMinPos;
    if Pos > HorMaxPos then Pos := HorMaxPos;
  end;

  Perform(WM_HSCROLL, (Pos shl 16) or SB_THUMBPOSITION, 0);
end;

// ---------------------------------------------------------------------------
// FetFormatLine -   
// ---------------------------------------------------------------------------
function TColorMemo.GetFormatLine(FmtRow: Longint): String;
var
  P: PFormatLine;
begin
  Result := '';
  if (FmtRow < 0) or (FmtRow > FFormatLines.Count - 1) then
    Exit;
  P := PFormatLine(FFormatLines.Items[FmtRow]);
  Result := P^.Line;
end;

// ---------------------------------------------------------------------------
// GetPaintBufferWidth -    .
// ---------------------------------------------------------------------------
function TColorMemo.GetPaintBufferWidth: Longint;
begin
  Result := ClientRect.Right;
  if FTextStatistic.TotalWidth > Result then
    Result := FTextStatistic.TotalWidth + 5; //  !
end;

// ---------------------------------------------------------------------------
// GetHorOffset -     .
// ---------------------------------------------------------------------------
function TColorMemo.GetHorOffset: Longint;
var
  HorOffset, PaintBufferWidth: Longint;
begin
  Result := 0;
  if msInvFormat in FMemoState then
    FormatText;
  if FTextStatistic.TotalWidth < ClientRect.Right then
    Exit;

  PaintBufferWidth := GetPaintBufferWidth;
  HorOffset := (FScrollState.HorPos - FScrollState.HorMinPos);
  HorOffset := Trunc(HorOffset* FTextStatistic.AvgTailCharW + 0.5);
  //     !
  if (HorOffset > PaintBufferWidth - ClientRect.Right) then
    HorOffset := PaintBufferWidth - ClientRect.Right;

  Result := HorOffset;
end;

// ---------------------------------------------------------------------------
// UpdateCaretYPos -  Y- 
// ---------------------------------------------------------------------------
procedure TColorMemo.UpdateCaretYPos;
var
  ScreenRow: Longint;
begin
  with FCaretState, FScrollState do begin
    if not CaretIsVertVisible then begin
      if FCaret.Visible then begin
        FCaret.Hide;
        Include(FMemoState, msCaretOutOfRange);
      end;
    end else begin
      ScreenRow := Row - (VertPos - VertMinPos);
      Y := TOP_MARGIN + ScreenRow*GetCaretHeight;
      FCaret.SetPos(X, Y);
      if msCaretOutOfRange in FMemoState then begin
        FCaret.Show;
        Exclude(FMemoState, msCaretOutOfRange);
      end;
    end;
  end;
end;

// ---------------------------------------------------------------------------
// CaretIsVertVisible -     ?
// ---------------------------------------------------------------------------
function TColorMemo.CaretIsVertVisible: Boolean;
begin
  Result := RowIsVisible(FCaretState.Row);
end;

// ---------------------------------------------------------------------------
// RowIsVisible -    ?
// ---------------------------------------------------------------------------
function TColorMemo.RowIsVisible(FmtRow: Longint): Boolean;
var
  ScreenRow: Longint;
begin
  with FTextStatistic, FScrollState do begin
    ScreenRow := FmtRow - (VertPos - VertMinPos);
    Result := (ScreenRow >= 0) and (ScreenRow < VertPageSize);
  end;
end;

// ---------------------------------------------------------------------------
// GetRowEndPos -  .     ( ).
// ---------------------------------------------------------------------------
function TColorMemo.GetRowEndPos(FmtRow: Longint; var CaretX: Longint): Longint;
var
  HorOffset: Longint;
  I: Longint;
  Line: String;
  CurX, CharW: Longint;
begin
  CaretX := LEFT_MARGIN;
  Result := 1;
  if FLines.Count = 0 then Exit;

  HorOffset := GetHorOffset;
  Line := GetFormatLine(FmtRow);
  CurX := LEFT_MARGIN;
  Result := Length(Line) + 1;
  for I := 1 to Length(Line) do begin
    CharW := Canvas.TextWidth(Line[I]);
    if (CurX + CharW) > (HorOffset + ClientRect.Right - RIGHT_MARGIN) then begin
      Result := I;
      break;
    end;
    Inc(CurX, CharW);
  end;
  CaretX := CurX - HorOffset;
end;

// ---------------------------------------------------------------------------
// PositionCaretXY -   (Row, Col)   
// . (X, Y) ()
// ---------------------------------------------------------------------------
procedure TColorMemo.PositionCaretXY(X, Y: Longint);
var
  I, TextX, LastX: Longint;
  H: Longint;
  HorOffset: Longint;
  Line: String;
  ScreenRow: Longint;
begin
  if FLines.Count = 0 then Exit;
  if (X < LEFT_MARGIN) or (X > ClientRect.Right - RIGHT_MARGIN) or
     (Y < TOP_MARGIN)  or (Y > ClientRect.Bottom - BOTTOM_MARGIN)
  then
    Exit;


  with FScrollState, FTextStatistic do begin
    HorOffset := GetHorOffset;
    Inc(X, HorOffset);

    H := GetCaretHeight;
    ScreenRow := (Y - TOP_MARGIN) div H;
    if ScreenRow > VisibleRows-1 then
      ScreenRow := VisibleRows-1;
    FCaretState.Row := VertPos - VertMinPos + ScreenRow;
    FCaretState.Col := GetRowEndPos(FCaretState.Row, FCaretState.X);
    FCaretState.Y := TOP_MARGIN + ScreenRow*H;

    Line := GetFormatLine(FCaretState.Row);
    LastX := LEFT_MARGIN;
    TextX := LEFT_MARGIN;
    for I := 1 to Length(Line) do begin
      TextX := TextX + Canvas.TextWidth(Line[I]);
      if TextX > X then begin
        if (X - LastX) > ((TextX - LastX) div 2) then begin
          FCaretState.X := TextX - HorOffset;
          FCaretState.Col := I + 1;
        end else begin
          FCaretState.X := LastX - HorOffset;
          FCaretState.Col := I;
        end;
        break;
      end;
      LastX := TextX;
    end;
  end;

  with FCaretState do
    FCaret.SetPos(X, Y);
end;

// ---------------------------------------------------------------------------
// UpdateCaretXPos -  X- 
// ---------------------------------------------------------------------------
procedure TColorMemo.UpdateCaretXPos;
var
  I: Longint;
  S: String;
  HorOffset: Longint;
begin
  with FCaretState do begin
    S := GetFormatLine(Row);
    HorOffset := GetHorOffset;
    X := LEFT_MARGIN;
    for I := 1 to Min(Col-1, Length(S)) do
      X := X + Canvas.TextWidth(S[I]);
    Dec(X, HorOffset);
    FCaret.SetPos(X, Y);
  end;
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.ClearState;
begin
  // FModified := False; //  ?
  ClearCaretState;
  ClearModifyState;
  ClearSelection;
  ClearScrollState;
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.SetLines(Value: TStrings);
begin
  ClearState;
  FLines.Assign(Value);
end;

// --------------------------------------------------------------------------
// TextChanged -   Lines.OnChange
// --------------------------------------------------------------------------
procedure TColorMemo.TextChanged(Sender: TObject);
begin
  if IsChanged and not (msTextChanging in FMemoState) then begin
    ClearState;
    InvalidateState;
    Repaint;
  end;
end;
// --------------------------------------------------------------------------
procedure TColorMemo.InvalidateState;
begin
  Include(FMemoState, msInvFormat);
  Include(FMemoState, msInvScrollRange);
  Include(FMemoState, msInvScrollPos);
end;

// --------------------------------------------------------------------------
procedure TColorMemo.SetScrollBars(Value: TScrollStyle);
begin
  if FScrollBars <> Value then begin
    FScrollBars := Value;
    UpdateAll(True);
  end;
end;

// --------------------------------------------------------------------------
// FormatText -   ( FormatLines)
// --------------------------------------------------------------------------
procedure TColorMemo.FormatText;
var
   OldFL: TFormatter;
   I: Longint;
   BegIndex, EndIndex: Longint;
   P: PFormatLine;
   Line: String;
begin
   if not Assigned(FFormatLines) then
     FFormatLines := CreateFormatLines;

   if FModifyState.FormatLines = nil then OldFL := FFormatLines
   else                                   OldFL := FModifyState.FormatLines;

   {$IFDEF DEBUG} DebugModifyFormatLines; {$ENDIF}

   FFormatLines := CreateFormatLines;

   with FFormatLines do
    if FScrollBars in [ssHorizontal, ssBoth] then
      Format(ftOneToOne)
    else if FWordWrap then
      Format(ftWordWrap)
    else
      Format(ftOneToMany);

   {$IFDEF DEBUG} DebugFormatLines; {$ENDIF}

   //  .     
   //  - GetPaintInfo
   BegIndex := 0;
   EndIndex := -1; //   !

   for I := 0 to FFormatLines.Count-1 do
      if I < OldFL.Count then begin
        Line := GetFormatLine(I);
        P := PFormatLine(OldFL.Items[I]);
        if  (Line = P^.Line) then begin
           if (EndIndex = -1) then
             BegIndex := I + 1;
        end else
          EndIndex := I;
      end else
        EndIndex := I;

   FFormatInfo.BegIndex := BegIndex;
   FFormatInfo.EndIndex := EndIndex;
   if EndIndex >= 0 then FFormatInfo.Valid := True
   else                  FFormatInfo.Valid := False; //  !


   GetTextStatistic;
   Exclude(FMemoState, msInvFormat);
end;

// ---------------------------------------------------------------------------
// Create - 
// ---------------------------------------------------------------------------
constructor TColorMemo.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  FVersion := COLOR_MEMO_VERSION;

  FLines := TStringList.Create;
  TStringList(FLines).OnChange := TextChanged;

  FFormatLines := nil;
  FModifyState.FormatLines := nil;

  FPaintLock := 0;
  FCaret := nil;

  ClearCaretState;
  ClearScrollState;
  ResetSelection;
  ClearModifyState;
  FShiftState := [];
  FMemoState := [msInvFormat, msInvScrollRange, msInvScrollPos, msPaintAll];

  FBorderStyle := bsSingle;
  FScrollBars := ssNone;
  Cursor := crIBeam;
  FReadOnly := False;
  FWordWrap := True;
  FHideSelection := True;
  FWantReturns := True;
  FMaxLength := 0;

  FOnGetSymbolColor := nil;
  FOldOnFontChange := Font.OnChange;
  Font.OnChange := FontChanged;

  FTextColor := TSymbolColor.Create(clBlack, clWhite);
  FTextColor.OnColorChange := TextColorChanged;
  FSelectColor := TSymbolColor.Create(clWhite, clHighlight);
  FSelectColor.OnColorChange := SelectColorChanged;
  FDisableColor := TSymbolColor.Create(clGray, clWhite);
  FDisableColor.OnColorChange := DisableColorChanged;

  TabStop := True;
  Width := CM_DEF_WIDTH;
  Height := CM_DEF_HEIGHT;
  ParentColor := CM_DEF_PARENTCOLOR;
end;

// ---------------------------------------------------------------------------
// Destroy - 
// ---------------------------------------------------------------------------
destructor TColorMemo.Destroy;
begin
  FLines.Free;
  FFormatLines.Free;

  if Assigned(FModifyState.FormatLines) then
    FModifyState.FormatLines.Free;
  if Assigned(FCaret) then
    FCaret.Free;

  FTextColor.Free;
  FSelectColor.Free;
  FDisableColor.Free;

  Font.OnChange := FOldOnFontChange;
  inherited Destroy;
end;

// ---------------------------------------------------------------------------
// CreateParams -   
// ---------------------------------------------------------------------------
procedure TColorMemo.CreateParams(var Params : TCreateParams);
begin
  inherited CreateParams(Params);

  Params.Style := Params.Style or BorderStyles[FBorderStyle];

{
  if FScrollBars in [ssVertical  , ssBoth] then
    Params.Style := Params.Style or WS_VSCROLL;
  if FScrollBars in [ssHorizontal, ssBoth] then
    Params.Style := Params.Style or WS_HSCROLL;
}
  Params.Style := Params.Style or WS_VSCROLL or WS_HSCROLL; 

  if NewStyleControls and Ctl3D and (FBorderStyle = bsSingle) then begin
    Params.Style := Params.Style and not WS_BORDER;
    Params.ExStyle := Params.ExStyle or WS_EX_CLIENTEDGE;
  end;
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.SetBorderStyle(Value: TBorderStyle);
begin
  if FBorderStyle <> Value then begin
    FBorderStyle := Value;
    RecreateWnd;
  end;
end;

// ---------------------------------------------------------------------------
// GetPaintInfo -   .
// ---------------------------------------------------------------------------
procedure TColorMemo.GetPaintInfo;
var
  TopIndex, BottomIndex: Longint;
  RowMin, ColMin, RowMax, ColMax: Longint;
  LineH: Longint;
begin
  LineH := Canvas.TextHeight('Test');
  GetTextStatistic;
  with FScrollState, FTextStatistic do begin
    TopIndex    := VertPos - VertMinPos;
    BottomIndex := TopIndex + VisibleRows - 1;
  end;

  with FPaintInfo, FFormatInfo, FSelectState do begin
    BegR := TopIndex;
    EndR := BottomIndex;
    if  (msPaintAll in FMemoState) or not (msRepaintWindow in FMemoState) then begin
      ;
    end else if FFormatInfo.Valid then  begin
      BegR := Max(TopIndex   , BegIndex);
      EndR := Min(BottomIndex, EndIndex);
    end else if IsSelected then begin
      GetSelMinMaxInfo(RowMin, RowMax, ColMin, ColMax);
      with FModifyState do begin
        RowMin := Min(SelectState.RowBeg, RowMin);
        RowMin := Min(SelectState.RowEnd, RowMin);
        RowMax := Max(SelectState.RowBeg, RowMax);
        RowMax := Max(SelectState.RowEnd, RowMax);
      end;
      BegR := Max(TopIndex   , RowMin);
      EndR := Min(BottomIndex, RowMax);
    end;
    Rect.Left := 0;
    Rect.Top  := 0;
    Rect.Right := GetPaintBufferWidth;
    Rect.Bottom := (EndR - BegR + 1)*LineH;
    TopFlag := False;
    if BegR = TopIndex then  begin
      Inc(Rect.Bottom, TOP_MARGIN);
      TopFlag := True;
    end;
    BottomFlag := False;
    if EndR  = BottomIndex then begin
      Rect.Bottom := ClientRect.Bottom; //  ,   !!!
      BottomFlag := True;
    end;

    TopY := 0;
    if not TopFlag then
      TopY := TOP_MARGIN + (BegR - TopIndex)*LineH;

    FFormatInfo.Valid := False;
    Exclude(FMemoState, msPaintAll);
  end;
end;

// ---------------------------------------------------------------------------
// Paint -  .
// ---------------------------------------------------------------------------
procedure TColorMemo.Paint;
var
  I, J, I0, J0: Longint;
  CurX, CurY: Longint;
  HorOffset: Longint;
  LineH, CharW: Longint;
  BufRect: TRect;
  ClipRect: TRect;
  hBmp  : HBITMAP;
  MemDC : HDC;
  cTextColor, cBkColor: TColor;
  OutLine: String;
  Select: Boolean;
begin

  if (msInvFormat in FMemoState) then
    FormatText;

  if (msInvScrollRange in FMemoState) then
    UpdateScrollRange;

  if (msInvScrollPos in FMemoState) then
    ResetScrollPos;

  if FPaintLock <> 0 then Exit;

  GetPaintInfo;
  //  ,    (. PaintNeeded)
  ClearModifyState;

  if Assigned(FOnStartPaint) then FOnStartPaint(Self);

  //
  // ---   ---
  //
  BufRect := FPaintInfo.Rect;

  HorOffset := 0;
  if FScrollBars in [ssHorizontal, ssBoth] then //   
    HorOffset := GetHorOffset;

  // ClipRect -   ,   
  //    
  ClipRect := ClientRect;
  ClipRect.Left  := ClipRect.Left  + HorOffset;
  ClipRect.Right := ClipRect.Right + HorOffset;

  MemDC := CreateCompatibleDC(Canvas.Handle);
  hBmp := CreateCompatibleBitmap(Canvas.Handle, BufRect.Right, BufRect.Bottom);


  SelectObject(MemDC, hBmp);
  SelectObject(MemDC, Canvas.Font.Handle);
  SetMapMode(MemDC, GetMapMode(Canvas.Handle));
  SetTextAlign(MemDC, TA_LEFT or TA_TOP);

  Canvas.Brush.Color := Color;
  FillRect(MemDC, BufRect,Canvas.Brush.Handle);

  if FPaintInfo.TopFlag then CurY := TOP_MARGIN
  else                       CurY := 0;

  LineH := Canvas.TextHeight('Test');

  for I := FPaintInfo.BegR to FPaintInfo.EndR do begin
    OutLine := GetFormatLine(I);

    if (OutLine <> '') then begin
      CurX := LEFT_MARGIN;
      for J := 1 to Length(OutLine) do begin
        CharW := Canvas.TextWidth(OutLine[J]);
        if CurX + CharW + RIGHT_MARGIN > BufRect.Right then
          break;

      Select := IsSelectedChar(I, J);
      if not Enabled then begin
        cTextColor := FDisableColor.TextColor;
        cBkColor   := FDisableColor.BackColor;
      end else if Select and (Focused or not FHideSelection) then begin
        cTextColor := FSelectColor.TextColor;;
        cBkColor   := FSelectColor.BackColor;
      end else begin
        cTextColor := FTextColor.TextColor;;
        cBkColor   := FTextColor.BackColor;
      end;

      if Assigned(FOnGetSymbolColor) then begin
        CaretPosFmtToReal(I0, J0, I, J);
        FOnGetSymbolColor(Self, OutLine[J], I0, J0, Select, cTextColor, cBkColor);
      end;

       Windows.SetTextColor(MemDC, ColorToRGB(cTextColor));
       Windows.SetBkColor(MemDC, ColorToRGB(cBkColor));
       ExtTextOut(MemDC, CurX, CurY, ETO_CLIPPED, @ClipRect, @OutLine[J], 1, nil);
       Inc(CurX, CharW);
     end;
   end;

   Inc(CurY, LineH);
  end;

  //    
  BitBlt(Canvas.Handle, 0, FPaintInfo.TopY, BufRect.Right, BufRect.Bottom,
        MemDC, HorOffset, 0, SrcCopy);

  DeleteDC(MemDC);
  DeleteObject(hBmp);

  if Assigned(FOnEndPaint) then FOnEndPaint(Self);
end;

// ---------------------------------------------------------------------------
// CMMAfterEnter -    .
// ---------------------------------------------------------------------------
procedure TColorMemo.CMMAfterEnter(var Msg : TMessage);
begin
  if Assigned(FAfterEnter) then
    FAfterEnter(Self);
end;

// ---------------------------------------------------------------------------
// CMMAfterExit -    .
// ---------------------------------------------------------------------------
procedure TColorMemo.CMMAfterExit(var Msg : TMessage);
begin
  if Assigned(FAfterExit) then
    FAfterExit(Self);
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.CMCtl3DChanged(var Msg : TMessage);
begin
  if (csLoading in ComponentState) then
    Exit;
  if not HandleAllocated then
    Exit;

  if NewStyleControls and (FBorderStyle = bsSingle) then
    RecreateWnd;

  inherited;
end;

// ---------------------------------------------------------------------------
//  CMEnabledChanged
//    - Enabled  Paint -  
//     .
// (. RepaintWindow  ,     !!!)
// ---------------------------------------------------------------------------
procedure TColorMemo.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  Repaint;
end;

// ---------------------------------------------------------------------------
// WMSetFont -     Font
// ---------------------------------------------------------------------------
procedure TColorMemo.WMSetFont(var Msg: TWMSetFont);
var
  RealRow, RealCol: Longint;
begin
  inherited;

  PaintLock;
  try
    with FCaretState do
      CaretPosFmtToReal(RealRow, RealCol, Row, Col);
    //         - Font!!!
    Canvas.Font := Font; // ,    WM_PAINT!
    //  ,   FTextColor    !
    FTextColor.TextColor := Font.Color;
    if Focused  and Assigned(FCaret) then
      FCaret.Height := Canvas.TextHeight('Caret');
    FormatText;
    UpdateScrollRange;
    with FCaretState do
      CaretPosRealToFmt(RealRow, RealCol, Row, Col);
    MakeCaretVisible;
    UpdateCaretYPos;
    UpdateCaretXPos;
  finally
    Include(FMemoState, msPaintAll);
    PaintUnlock;
  end;
end;

// ---------------------------------------------------------------------------
//  WMRButtonUp -   
// ---------------------------------------------------------------------------
procedure TColorMemo.WMRButtonUp(var Msg : TWMRButtonDown);
var
  P  : TPoint;
  M  : TPopUpMenu;
  MI : TMenuItem;
begin
  inherited;
  if csDesigning in ComponentState then
    Exit;

  if not Focused and CanFocus then
    SetFocus;

  inherited;
  if PopUpMenu = nil then begin
    M := TPopupMenu.Create(Self);
    try
      MI := TMenuItem.Create(M);
      MI.Caption := MI_UNDO_TEXT;
      MI.Enabled := FUndoBuffer.Valid;
      MI.OnClick := UndoClick;
      M.Items.Add(MI);

      MI := TMenuItem.Create(M);
      MI.Caption := '-';
      M.Items.Add(MI);

      MI := TMenuItem.Create(M);
      MI.Caption := MI_CUT_TEXT;
      MI.Enabled := IsSelected and not FReadOnly;
      MI.OnClick := CutClick;
      M.Items.Add(MI);

      MI := TMenuItem.Create(M);
      MI.Caption := MI_COPY_TEXT;
      MI.Enabled := IsSelected;
      MI.OnClick := CopyClick;
      M.Items.Add(MI);

      MI := TMenuItem.Create(M);
      MI.Caption := MI_PASTE_TEXT;
      MI.Enabled := not FReadOnly and Clipboard.HasFormat(CF_TEXT);
      MI.OnClick := PasteClick;
      M.Items.Add(MI);

      MI := TMenuItem.Create(M);
      MI.Caption := MI_DELETE_TEXT;
      MI.Enabled := IsSelected and not FReadOnly;
      MI.OnClick := DeleteClick;
      M.Items.Add(MI);

      MI := TMenuItem.Create(M);
      MI.Caption := '-';
      M.Items.Add(MI);

      MI := TMenuItem.Create(M);
      MI.Caption := MI_SELECT_ALL_TEXT;
      MI.Enabled := not IsAllSelected;
      MI.OnClick := SelectAllClick;
      M.Items.Add(MI);

      P.X := Msg.XPos;
      P.Y := Msg.YPos;
      P := ClientToScreen(P);
      M.PopUp(P.X, P.Y);

      Application.ProcessMessages;
    finally
      M.Free;
    end;
  end;
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.WMCopy(var Msg : TWMCopy);
begin
  inherited;
  CmdCopy;
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.WMCut(var Msg : TWMCut);
begin
  inherited;
  CmdCut;
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.WMPaste(var Msg : TWMPaste);
begin
  inherited;
  CmdPaste;
end;

// ---------------------------------------------------------------------------
// WMChar -  .
// ---------------------------------------------------------------------------
procedure TColorMemo.WMChar(var Msg : TWMChar);
begin
  inherited;
  if csDesigning in ComponentState then
    Exit;

  if ssCtrl in FShiftState then begin
    case Msg.CharCode of
      CMD_COPY : CmdCopy;
      CMD_CUT  : CmdCut;
      CMD_PASTE: CmdPaste;
      CMD_UNDO : CmdUndo;
    end;
    Exit;
  end;

  CmdChar(Msg.CharCode);
end;

// ---------------------------------------------------------------------------
//  WMSysKeyDown -     <Alt+BackSpace>
// ---------------------------------------------------------------------------
procedure TColorMemo.WMSysKeyDown(var Msg : TWMSysKeyDown);
begin
  inherited;

  if (csDesigning in ComponentState) or FReadOnly then
    Exit;
  FShiftState := KeyDataToShiftState(Msg.KeyData);
  if (Msg.CharCode = VK_BACK) and (ssAlt in FShiftState) then
     CmdUndo;
end;

// ---------------------------------------------------------------------------
// WMKeyDown -  .
// ---------------------------------------------------------------------------
procedure TColorMemo.WMKeyDown(var Msg : TWMKeyDown);
var
  ParentForm: TForm;
begin
  inherited;

  if (csDesigning in ComponentState) then
    Exit;

  FShiftState := KeyDataToShiftState(Msg.KeyData);

  case Msg.CharCode of
    VK_LEFT:    CmdGoLeft;
    VK_RIGHT:   CmdGoRight;
    VK_UP:      CmdGoUp;
    VK_PRIOR:   CmdGoPgUp;
    VK_DOWN:    CmdGoDown;
    VK_NEXT:    CmdGoPgDown;
    VK_HOME:
      if ssCtrl in FShiftState then CmdGoStart
      else                          CmdGoHome;
    VK_END:
      if ssCtrl in FShiftState then CmdGoFinish
      else                          CmdGoEnd;
    VK_RETURN:
      if (ssCtrl in FShiftState) or FWantReturns then
        CmdNewLine
      else begin //     
        ParentForm := TForm(GetParentForm(Self));
        if ParentForm <> nil then
          ParentForm.Perform(WM_NEXTDLGCTL, 0, 0);
      end;
    VK_DELETE:
      if IsSelected then  CmdDeleteSelected
      else                CmdDeleteChar( 1);
    VK_BACK:
      if IsSelected then  CmdDeleteSelected
      else                CmdDeleteChar(-1);
    VK_INSERT: begin
      if      ssCtrl  in FShiftState then CmdCopy
      else if ssShift in FShiftState then CmdPaste;
    end;
  end;
end;

// ---------------------------------------------------------------------------
// WMSetFocus -   .
// ---------------------------------------------------------------------------
procedure TColorMemo.WMSetFocus(var Msg : TWMSetFocus);
begin
  inherited;

  FCaret := TCaret.Create(Handle, 2*GetSystemMetrics(SM_CYBORDER), GetCaretHeight);
  with FCaretState do
    FCaret.SetPos(X, Y);
  FCaret.Show;

  Repaint;

  PostMessage(Handle, CMM_AFTERENTER, 0, 0); //    AfterEnter
end;

// ---------------------------------------------------------------------------
// WMKillFocus -  .
// ---------------------------------------------------------------------------
procedure TColorMemo.WMKillFocus(var Msg : TWMKillFocus);
begin
  inherited;

  Repaint;

  if MouseCapture then MouseCapture := False;
  Exclude(FMemoState, msMouseSelection);

  if Assigned(FCaret) then
    FCaret.Free;
  FCaret := nil;

  PostMessage(Handle, CMM_AFTEREXIT, 0, 0); //    AfterExit
end;

// ---------------------------------------------------------------------------
// WMLButtonDown -    .
// ---------------------------------------------------------------------------
procedure TColorMemo.WMLButtonDown(var Msg : TWMLButtonDown);
var
  OldCaretRow, OldCaretCol: Longint;
begin
  inherited;

  if csDesigning in ComponentState then Exit;

  FShiftState := KeysToShiftState(Msg.Keys);

  OldCaretRow := FCaretState.Row;
  OldCaretCol := FCaretState.Col;

  PositionCaretXY(Msg.XPos, Msg.YPos);

  if InSelectMode then begin
    PerformSelection(OldCaretRow, OldCaretCol);
    RepaintWindow;
  end else if IsSelected then begin
    ResetSelection;
    RepaintWindow;
  end;

  Include(FMemoState, msMouseSelection);

  if not Focused then SetFocus;

  FCaret.Show;
end;

// ---------------------------------------------------------------------------
// WMLButtonUp -    
// ---------------------------------------------------------------------------
procedure TColorMemo.WMLButtonUp(var Msg : TWMLButtonUp);
begin
  inherited;

  Exclude(FMemoState, msMouseSelection);
end;

// ---------------------------------------------------------------------------
// WMMouseMove -  .
// ---------------------------------------------------------------------------
procedure TColorMemo.WMMouseMove(var Msg : TWMMouseMove);
var
  OldCaretRow, OldCaretCol: Longint;
begin
  inherited;

  if csDesigning in ComponentState then
    Exit;
  if not Focused then
    Exit;

  if msMouseSelection in FMemoState then begin
    if Msg.XPos < LEFT_MARGIN then
      CmdGoLeft
    else if Msg.XPos > (ClientRect.Right - RIGHT_MARGIN) then
      CmdGoRight
    else if Msg.YPos < TOP_MARGIN then
      CmdGoUp
    else if Msg.YPos > (ClientRect.Bottom - BOTTOM_MARGIN) then
      CmdGoDown
    else begin
      OldCaretRow := FCaretState.Row;
      OldCaretCol := FCaretState.Col;
      PositionCaretXY(Msg.XPos, Msg.YPos);
      if (FCaretState.Row <> OldCaretRow) or (FCaretState.Col <> OldCaretCol) then
      begin
         PerformSelection(OldCaretRow, OldCaretCol);
         RepaintWindow;
         FCaret.Show;
      end;
    end;
  end;
end;

// ---------------------------------------------------------------------------
// WMWindowPosChanged -   !
//    Width  Height.
// ---------------------------------------------------------------------------
procedure TColorMemo.WMWindowPosChanged(var Message: TWMWindowPosChanged);
var
  RealRow, RealCol: Longint;
begin
  inherited;
  if (csDestroying in ComponentState) then Exit;

  //      UpdateScrollRange 
  //   -,   - 
  //  !!!
  //      (msInv...),  
  //   UpdateScrollRange    !!!
  with FCaretState do
    CaretPosFmtToReal(RealRow, RealCol, Row, Col);

  FormatText;
  UpdateScrollRange;

  with FCaretState do
    CaretPosRealToFmt(RealRow, RealCol, Row, Col);

  UpdateCaretYPos;
  UpdateCaretXPos;  
end;

// ---------------------------------------------------------------------------
// WMEraseBkGnd -     WM_PAINT  
//    ,    .
// ---------------------------------------------------------------------------
procedure TColorMemo.WMEraseBkGnd(var Msg : TWMEraseBkGnd);
begin
  Msg.Result := 1;  //       !!!
end;

// ---------------------------------------------------------------------------
// WMVScroll -  .
// ---------------------------------------------------------------------------
procedure TColorMemo.WMVScroll(var Msg: TWMVScroll);
  procedure DoScroll;
  begin
    SetScrollPos(Handle, SB_VERT, FScrollState.VertPos, True);
    Include(FMemoState, msPaintAll);
    Repaint;
    UpdateCaretYPos;
  end;
begin
  inherited;

  with FScrollState do
    case Msg.ScrollCode of
      SB_LINEUP: begin
        if  VertPos > VertMinPos then begin
          Dec(VertPos);
          DoScroll;
        end;
      end;
      SB_LINEDOWN: begin
        if VertPos < VertMaxPos then begin
          Inc(VertPos);
          DoScroll;
        end;
      end;
      SB_THUMBPOSITION: begin
        VertPos := Msg.Pos;
        DoScroll;
      end;
      SB_PAGEUP: begin
        Dec(VertPos, VERT_SCROLL_CHUNK);
        if VertPos < VertMinPos then
          VertPos := VertMinPos;
        DoScroll;
      end;
      SB_PAGEDOWN: begin
        Inc(VertPos, VERT_SCROLL_CHUNK);
        if VertPos > VertMaxPos then
          VertPos := VertMaxPos;
        DoScroll;
      end;
    end;
end;

// ---------------------------------------------------------------------------
// WMHScroll -  .
// ---------------------------------------------------------------------------
procedure TColorMemo.WMHScroll(var Msg: TWMHScroll);
  procedure DoScroll;
  begin
    SetScrollPos(Handle, SB_HORZ, FScrollState.HorPos, True);
    Include(FMemoState, msPaintAll);
    Repaint;
    UpdateCaretXPos;
  end;
begin
  inherited;

  with FScrollState do
    case Msg.ScrollCode of
      SB_LINEUP: begin
        if  HorPos > HorMinPos then begin
          Dec(HorPos);
          DoScroll;
        end;
      end;
      SB_LINEDOWN: begin
        if HorPos < HorMaxPos then begin
          Inc(HorPos);
          DoScroll;
        end;
      end;
      SB_THUMBPOSITION: begin
        HorPos := Msg.Pos;
        DoScroll;
      end;
      SB_PAGEUP: begin
        Dec(HorPos, HOR_SCROLL_CHUNK);
        if HorPos < HorMinPos then
          HorPos := HorMinPos;
        DoScroll;
      end;
      SB_PAGEDOWN: begin
        Inc(HorPos, HOR_SCROLL_CHUNK);
        if HorPos > HorMaxPos then
          HorPos := HorMaxPos;
        DoScroll;
      end;
    end;
end;

// ---------------------------------------------------------------------------
// WMLButtonDblClk -   .
// ---------------------------------------------------------------------------
procedure TColorMemo.WMLButtonDblClk(var Msg : TWMLButtonDblClk);
begin
  inherited;

  if csDesigning in ComponentState then Exit;
  if FLines.Count = 0 then Exit;

  CmdSelectWord;
end;

// ---------------------------------------------------------------------------
// GetTextStatistic -    FTextStatistic   .
// ---------------------------------------------------------------------------
procedure  TColorMemo.GetTextStatistic;
  function GetPageInfo(var VertRows: Longint): Longint;
  var
    I: Longint;
    LineH: Longint;
    CurX: Longint;
  begin
    Result := 0;
    VertRows := 0;
    CurX := TOP_MARGIN + BOTTOM_MARGIN;
    LineH := Canvas.TextHeight('Test');
    I := GetFirstRow;
    while True do begin
      Inc(CurX, LineH);
      if CurX <= ClientRect.Bottom then Inc(Result)
      else                              break;
      if I < FFormatLines.Count then
        Inc(VertRows);
      Inc(I);
    end;
  end;
var
  I, IMax: Longint;
  StrW, CharW: Longint;
  CurX: Longint;
begin
  with FTextStatistic do begin
    TotalWidth := ClientRect.Right;
    AvgTailCharW := 0;
    TailCharCount := 0;
    VertPageSize := GetPageInfo(VisibleRows);

    if FScrollBars = ssVertical then
      Exit;
    if FLines.Count = 0 then
      Exit;

    TotalWidth := LEFT_MARGIN + RIGHT_MARGIN;
    IMax := -1;
    with FLines do
      for I := 0 to Count-1 do begin
        StrW := Canvas.TextWidth(FLines[I]);
        if  TotalWidth < StrW then begin
          TotalWidth := StrW;
        IMax := I;
      end;
    end;

    if IMax < 0 then Exit;

    CurX := LEFT_MARGIN;
    VisibleChars := 0;
    for I := 1 to Length(FLines[IMax]) do begin
      CharW := Canvas.TextWidth(FLines[IMax][I]);
      if CurX + CharW + RIGHT_MARGIN > ClientRect.Right then
        break;
      Inc(VisibleChars);
      Inc(CurX, CharW);
    end;
    TailCharCount := Length(FLines[IMax]) - VisibleChars;
    if TailCharCount > 0 then
      AvgTailCharW := (TotalWidth - ClientRect.Right)/TailCharCount;
  end;

  {$IFDEF DEBUG} DebugTextStatistic; {$ENDIF}
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.ResetScrollPos;
begin
  with FScrollState do begin
    VertPos := 1;
    HorPos := 1;
  end;

  UpdateScrollPos;
end;

// ---------------------------------------------------------------------------
procedure TColorMemo.UpdateScrollPos;
begin
  if (msInvScrollRange in FMemoState) then
    UpdateScrollRange;

  with FScrollState do begin
    if VertPos < VertMinPos then VertPos := VertMinPos;
    if VertPos > VertMaxPos then VertPos := VertMaxPos;
    if HorPos < HorMinPos then   HorPos := HorMinPos;
    if HorPos > HorMaxPos then   HorPos := HorMaxPos;
    SetScrollPos(Handle, SB_VERT, VertPos, True);
    SetScrollPos(Handle, SB_HORZ, HorPos, True);
  end;

  Exclude(FMemoState, msInvScrollPos);
end;

// ---------------------------------------------------------------------------
// UpdateScrollRange -    .  . 
// ---------------------------------------------------------------------------
procedure TColorMemo.UpdateScrollRange;
var
  ScrollInfo: TScrollInfo;
  OldClientRect: TRect;

  function IsRectEqual(const A, B: TRect): Boolean;
  begin
    Result := (A.Left = B.Left) and (A.Right = B.Right) and
              (A.Top = B.Top)   and (A.Bottom = B.Bottom)
  end;
begin
  if not HandleAllocated then
    Exit;

  if (msInvFormat in FMemoState) then
    FormatText;

  OldClientRect := ClientRect;

  FScrollState.VertMinPos := 1;
  with FTextStatistic do
    if FFormatLines.Count <= VertPageSize then
      FScrollState.VertMaxPos := FScrollState.VertMinPos
    else begin
      if VisibleRows < VertPageSize then begin
        VisibleRows := VertPageSize;
        //     ( !)
        Include(FMemoState, msPaintAll);
      end;
      FScrollState.VertMaxPos := FFormatLines.Count - FTextStatistic.VisibleRows + 1;
    end;

  if FScrollBars in [ssVertical, ssBoth] then begin
    with FScrollState, FTextStatistic, ScrollInfo do begin
      cbSize := sizeof(ScrollInfo);
      fMask := SIF_PAGE or SIF_RANGE;
      nMin := VertMinPos;
      nMax := FFormatLines.Count;
      nPage := VertPageSize;

      SetScrollInfo(Handle, SB_VERT, ScrollInfo, True);
    end;
  end else
    SetScrollRange(Handle, SB_VERT, 1, 1, True);

  // :  nMax  VertMaxPos  !!!   !
  //             VertMaxPos      nPage, 
  //              WMVScroll   !

  if FScrollBars in [ssHorizontal, ssBoth] then begin

    FScrollState.HorMinPos := 1;
    if ClientRect.Right >= FTextStatistic.TotalWidth then
      FScrollState.HorMaxPos := FScrollState.HorMinPos
    else begin
      FScrollState.HorMaxPos := FTextStatistic.TailCharCount + 2;
    end;

    with FScrollState, FTextStatistic, ScrollInfo do begin
      cbSize := sizeof(ScrollInfo);
      fMask := SIF_PAGE or SIF_RANGE;
      nMin := HorMinPos;
      nMax := VisibleChars + TailCharCount;
      nPage := VisibleChars;

      SetScrollInfo(Handle, SB_HORZ, ScrollInfo, True);
    end;
  end else
      SetScrollRange(Handle, SB_HORZ, 1, 1, True);

  Exclude(FMemoState, msInvScrollRange);


  UpdateScrollPos;
  //   - ScrollBar  
  //  
  if not IsRectEqual(OldClientRect, ClientRect) then begin
    //    (PaintLock),   Repaint  !
    //         msRepaintWindow,
    //      Repaint   RepaintWindow   ,
    //      (PaintUnlock) 
    //  - RepaintWindow!!!    msPaintAll,
    //        RepaintWindow.
    Include(FMemoState, msPaintAll);
    Repaint;
  end;

  {$IFDEF DEBUG_SCROLL_RANGE} DebugScrollRange; {$ENDIF}
end;

// ---------------------------------------------------------------------------
// WMSize -   .
// -     (. WMWindowPosChange)
// ---------------------------------------------------------------------------
procedure TColorMemo.WMSize(var Msg: TWMSize);
begin
  inherited;
  InvalidateState;
end;

// ---------------------------------------------------------------------------
//  WMGetDlgCode -   !!!
//       ,   <Left. Right, Up, Down>
//        -  ,   
//    <Tab>!!!
// ---------------------------------------------------------------------------
procedure TColorMemo.WMGetDlgCode(var Msg : TWMGetDlgCode);
begin
  inherited;

  if csDesigning in ComponentState then
    Msg.Result := DLGC_STATIC
  else
    Msg.Result := Msg.Result or DLGC_WANTCHARS or DLGC_WANTARROWS;
end;

end.
