// ===========================================================================
//                                                                          //
//  TDBColorEdit  1.01 (C)  .. 1998                           //
//                                                                          //
// ===========================================================================

unit DBColEdt;

interface

uses
 Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
 StdCtrls, ExtCtrls, DB, DBCtrls, ColorEdt;

type
  TDBColorEdit = class(TColorEdit)
  private
    FDataLink: TFieldDataLink;
    FTextLoaded: Boolean;

    function  GetDataField: string;
    function  GetDataSource: TDataSource;
    procedure SetDataField(const Value: string);
    procedure SetDataSource(Value: TDataSource);
    procedure DataChange(Sender: TObject);
    procedure UpdateData(Sender: TObject);
    procedure CMGetDataLink(var Msg: TMessage); message CM_GETDATALINK;
    function  GetField: TField;
    function  GetReadOnly: Boolean;
    procedure SetReadOnly(Value: Boolean);
    procedure EditingChange(Sender: TObject);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure Change; override;
    property  TextLoaded: Boolean read FTextLoaded;
    procedure KeyDown (var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure WMPaste(var Msg : TWMPaste); message WM_PASTE;
    procedure WMCut(var Msg : TWMCut); message WM_CUT;
    procedure CMExit(var Msg: TCMExit); message CM_EXIT;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Field: TField read GetField;
  published
    property DataField: String
             read GetDataField  write SetDataField;
    property DataSource: TDataSource
             read GetDataSource write SetDataSource;
    // - ReadOnly    ,    ,
    //           
    property ReadOnly: Boolean
             read GetReadOnly   write SetReadOnly default False;
  end;

implementation

const
 DB_COLOR_EDIT_VERSION = '1.01';

{TDBColorEdit}

// ---------------------------------------------------------------------------
function  TDBColorEdit.GetField: TField;
begin
  Result := FDataLink.Field;
end;

// ---------------------------------------------------------------------------
function  TDBColorEdit.GetDataField: string;
begin
  Result := FDataLink.FieldName;
end;

// ---------------------------------------------------------------------------
function  TDBColorEdit.GetDataSource: TDataSource;
begin
  Result := FDataLink.DataSource;
end;

// ---------------------------------------------------------------------------
procedure TDBColorEdit.SetDataField(const Value: string);
begin
  FDataLink.FieldName := Value;
  if FDataLink.Field = nil then MaxLength := 0
  else                          MaxLength := FDataLink.Field.Size;
end;

// ---------------------------------------------------------------------------
// SetDataSource
//  FreeNotification   .   Value 
//   (Self)   ,   
//     Value.   Value 
//    ,   Self,   ( Notification) 
//   FreeNotification  . ,    
//  Value  DataSource     
//  ,    FreeNotification !!!
// ---------------------------------------------------------------------------
procedure TDBColorEdit.SetDataSource(Value: TDataSource);
begin
  FDataLink.DataSource := Value;
  if Value <> nil then Value.FreeNotification(Self);
end;

// ---------------------------------------------------------------------------
// DataChange -      
//    - Text    :
// (Field.Text, Field.DisplayText).   ,     
//        
// ---------------------------------------------------------------------------
procedure TDBColorEdit.DataChange(Sender: TObject);
begin
  FTextLoaded := False;
  if FDataLink.Field = nil then
    Clear
  else if FDataLink.CanModify then begin
    Text := FDataLink.Field.Text;
    FTextLoaded := True;
  end else begin
    Text := FDataLink.Field.DisplayText;
    FTextLoaded := True;
  end;
  Modified := False;
end;

// ---------------------------------------------------------------------------
//  UpdateData      
// ---------------------------------------------------------------------------
procedure TDBColorEdit.UpdateData(Sender: TObject);
begin
  FDataLink.Field.AsString := Text;
end;

// ---------------------------------------------------------------------------
// CMGetDataLink -   -   ( )
// ---------------------------------------------------------------------------
procedure TDBColorEdit.CMGetDataLink(var Msg: TMessage);
begin
  Msg.Result := Longint(FDataLink);
end;

// ---------------------------------------------------------------------------
// Notification -        
// ---------------------------------------------------------------------------
procedure TDBColorEdit.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);

  if (FDataLink <> nil) and
     (AComponent = DataSource) and
     (Operation = opRemove)
  then
    DataSource := nil;
end;

// ---------------------------------------------------------------------------
constructor TDBColorEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  inherited ReadOnly := True; // ReadOnly !
  ControlStyle := ControlStyle + [csReplicatable]; //  TDBCtrlGrid

  FVersion := DB_COLOR_EDIT_VERSION;
  FTextLoaded := False;
  FDataLink := TFieldDataLink.Create;
  //   
  with FDataLink do begin
    Control := Self;
    OnDataChange := DataChange;
    OnUpdateData := UpdateData;
    OnEditingChange := EditingChange;
  end;
end;

// ---------------------------------------------------------------------------
destructor TDBColorEdit.Destroy;
begin
  FDataLink.Free;

  inherited Destroy;
end;

// ---------------------------------------------------------------------------
procedure TDBColorEdit.Change;
begin
  if TextLoaded then FDataLink.Modified;
  inherited Change;
end;

// ---------------------------------------------------------------------------
function  TDBColorEdit.GetReadOnly: Boolean;
begin
  Result := FDataLink.ReadOnly;
end;

// ---------------------------------------------------------------------------
procedure TDBColorEdit.SetReadOnly(Value: Boolean);
begin
  FDataLink.ReadOnly := Value;
end;


// ---------------------------------------------------------------------------
procedure TDBColorEdit.EditingChange(Sender: TObject);
begin
  inherited ReadOnly := not (FDataLink.Editing and TextLoaded);
end;

// ---------------------------------------------------------------------------
procedure TDBColorEdit.KeyDown (var Key: Word; Shift: TShiftState);
begin
  inherited KeyDown(Key, Shift);
  if TextLoaded then begin
    if Key in [VK_DELETE, VK_BACK, VK_CLEAR] then
      FDataLink.Edit;
  end else
    Key := 0;
end;

// ---------------------------------------------------------------------------
procedure TDBColorEdit.KeyPress(var Key: Char);
begin
  inherited KeyPress(Key);

  if TextLoaded then begin
    if (Key in [#32..#255]) and (FDataLink.Field <> nil) and
       (not FDataLink.Field.IsValidChar(Key) or
       (Length(Text) >=  FDataLink.Field.Size)) then
    begin
      MessageBeep(Cardinal(-1));
      Key := #0;
    end;

    case Key of
      ^I, ^J, ^M, #32..#255:
         FDataLink.Edit;
      #27:
         FDataLink.Reset;
    end;
  end else
    Key := #0;
end;

// ---------------------------------------------------------------------------
procedure TDBColorEdit.WMPaste(var Msg: TWMPaste);
begin
  if TextLoaded then begin
    FDataLink.Edit;
    inherited;
  end;
end;

// ---------------------------------------------------------------------------
procedure TDBColorEdit.WMCut(var Msg: TWMCut);
begin
  if TextLoaded then begin
    FDataLink.Edit;
    inherited;
  end;
end;

// ---------------------------------------------------------------------------
procedure TDBColorEdit.CMExit(var Msg: TCMExit);
begin
  try
    if FDataLink.Editing then
      FDataLink.UpdateRecord;
  except
    SelectAll;
    SetFocus;
    raise;
  end;

  inherited;
end;

end. // DBColorEd
