unit DBColMem;

// ===========================================================================
//                                                                          //
//  TDBColorMemo  1.01 (C)  .. 1998                           //
//                                                                          //
// ===========================================================================

interface

uses
 Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
 StdCtrls, ExtCtrls, DB, DBTables, DBCtrls, ColorMem;

type
  TDBColorMemo = class(TColorMemo)
  private
    FDataLink: TFieldDataLink;
    FTextLoaded: Boolean;
    FUpdatingData: Boolean;
    FAutoDisplay: Boolean;

    function  GetDataField: string;
    function  GetDataSource: TDataSource;
    procedure SetDataField(const Value: string);
    procedure SetDataSource(Value: TDataSource);
    procedure DataChange(Sender: TObject);
    procedure UpdateData(Sender: TObject);
    procedure CMGetDataLink(var Msg: TMessage); message CM_GETDATALINK;
    function  GetField: TField;
    function  GetReadOnly: Boolean;
    procedure SetReadOnly(Value: Boolean);
    procedure EditingChange(Sender: TObject);
    procedure SetAutoDisplay(Value: Boolean);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure Change; override;
    procedure KeyDown (var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure WMPaste(var Msg : TWMPaste); message WM_PASTE;
    procedure WMCut(var Msg : TWMCut); message WM_CUT;
    procedure WMLButtonDblClk(var Msg: TWMLButtonDblClk); message WM_LBUTTONDBLCLK;
    procedure CMExit(var Msg: TCMExit); message CM_EXIT;
    procedure LoadText; virtual;
    property  TextLoaded: Boolean read FTextLoaded;
    property  UpdatingData: Boolean read FUpdatingData;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Field: TField read GetField;
  published
    property AutoDisplay: Boolean
             read FAutoDisplay write SetAutoDisplay default True;
    property DataField: String
             read GetDataField  write SetDataField;
    property DataSource: TDataSource
             read GetDataSource write SetDataSource;
    // - ReadOnly    ,    ,
    //           
    property ReadOnly: Boolean
             read GetReadOnly   write SetReadOnly default False;
  end;

implementation

const
 DB_COLOR_MEMO_VERSION = '1.01';

{TDBColorMemo}


// ---------------------------------------------------------------------------
procedure TDBColorMemo.SetAutoDisplay(Value: Boolean);
begin
  if Value <> FAutoDisplay then
  begin
    FAutoDisplay := Value;
    if Value then LoadText;
  end;
end;

// ---------------------------------------------------------------------------
function  TDBColorMemo.GetField: TField;
begin
  Result := FDataLink.Field;
end;

// ---------------------------------------------------------------------------
function  TDBColorMemo.GetDataField: string;
begin
  Result := FDataLink.FieldName;
end;

// ---------------------------------------------------------------------------
function  TDBColorMemo.GetDataSource: TDataSource;
begin
  Result := FDataLink.DataSource;
end;

// ---------------------------------------------------------------------------
procedure TDBColorMemo.SetDataField(const Value: string);
begin
  FDataLink.FieldName := Value;
  if FDataLink.Field = nil then MaxLength := 0
  else                          MaxLength := FDataLink.Field.Size;
end;

// ---------------------------------------------------------------------------
// SetDataSource
//  FreeNotification   .   Value 
//   (Self)   ,   
//     Value.   Value 
//    ,   Self,   ( Notification) 
//   FreeNotification  . ,    
//  Value -  DataSource     
//  ,    FreeNotification !!!
// ---------------------------------------------------------------------------
procedure TDBColorMemo.SetDataSource(Value: TDataSource);
begin
  FDataLink.DataSource := Value;
  if Value <> nil then Value.FreeNotification(Self);
end;

// ---------------------------------------------------------------------------
// DataChange -      
//    - Text    :
// (Field.Text, Field.DisplayText).   ,     
//        
// ---------------------------------------------------------------------------
procedure TDBColorMemo.DataChange(Sender: TObject);
begin
  if FDataLink.Field = nil then
    Clear
  else if FDataLink.Field is TBlobField then begin
    //  BLOB-
    if AutoDisplay or (FDataLink.Editing and TextLoaded) then begin
      //     AutoDisplay=True,   
      //  
      FTextLoaded := False;
      LoadText;
    end else begin
      //       , 
      //    
      Lines.Text := '(' + FDataLink.Field.DisplayLabel + ')';
      FTextLoaded := False;
    end;
  end else if FDataLink.CanModify then begin
    //   BLOB-,     
    // ( )
    Lines.Text := FDataLink.Field.Text;
    FTextLoaded := True;
  end else begin
    // -//-
    // ( )
    Lines.Text := FDataLink.Field.DisplayText;
    FTextLoaded := True;
  end;

  Modified := False;
end;

// ---------------------------------------------------------------------------
// LoadText -     
// ---------------------------------------------------------------------------
procedure TDBColorMemo.LoadText;
var
  BlobStream: TBlobStream;
begin
  if TextLoaded or (FDataLink.Field = nil) or
     not (FDataLink.Field is TBlobField)
  then
    Exit;

  FUpdatingData := True;
  try
    BlobStream := TBlobStream.Create(TBlobField(FDataLink.Field), bmRead);
    try
      Lines.LoadFromStream(BlobStream);
    finally
      BlobStream.Free;
    end;
  finally
    FUpdatingData := False;
  end;

  FTextLoaded := True;
end;

// ---------------------------------------------------------------------------
//  UpdateData      
// ---------------------------------------------------------------------------
procedure TDBColorMemo.UpdateData(Sender: TObject);
var
  BlobStream: TBlobStream;
begin
  if FDataLink.Field is TBlobField then begin
    BlobStream := TBlobStream.Create(TBlobField(FDataLink.Field), bmWrite);
    try
      Lines.SaveToStream(BlobStream);
    finally
      BlobStream.Free;
    end;
  end else
    FDataLink.Field.AsString := Text;
end;

// ---------------------------------------------------------------------------
// CMGetDataLink -   -   ( )
// ---------------------------------------------------------------------------
procedure TDBColorMemo.CMGetDataLink(var Msg: TMessage);
begin
  Msg.Result := Longint(FDataLink);
end;

// ---------------------------------------------------------------------------
// Notification -        
// ---------------------------------------------------------------------------
procedure TDBColorMemo.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);

  if (FDataLink <> nil) and
     (AComponent = DataSource) and
     (Operation = opRemove)
  then
    DataSource := nil;
end;

// ---------------------------------------------------------------------------
constructor TDBColorMemo.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  inherited ReadOnly := True; // ReadOnly (TColorMemo)!
  ControlStyle := ControlStyle + [csReplicatable];  //  TDBCtrlGrid

  FVersion := DB_COLOR_MEMO_VERSION;
  FTextLoaded := False;
  FUpdatingData := False;
  FAutoDisplay := True;
  FDataLink := TFieldDataLink.Create;
  //   
  with FDataLink do begin
    Control := Self;
    OnDataChange := DataChange;
    OnUpdateData := UpdateData;
    OnEditingChange := EditingChange;
  end;
end;

// ---------------------------------------------------------------------------
destructor TDBColorMemo.Destroy;
begin
  FDataLink.Free;

  inherited Destroy;
end;

// ---------------------------------------------------------------------------
procedure TDBColorMemo.Change;
begin
  if TextLoaded then FDataLink.Modified;
  inherited Change;
end;

// ---------------------------------------------------------------------------
function  TDBColorMemo.GetReadOnly: Boolean;
begin
  Result := FDataLink.ReadOnly;
end;

// ---------------------------------------------------------------------------
procedure TDBColorMemo.SetReadOnly(Value: Boolean);
begin
  FDataLink.ReadOnly := Value;
end;

// ---------------------------------------------------------------------------
procedure TDBColorMemo.EditingChange(Sender: TObject);
begin
  inherited ReadOnly := not (FDataLink.Editing and TextLoaded);
end;

// ---------------------------------------------------------------------------
procedure TDBColorMemo.KeyDown (var Key: Word; Shift: TShiftState);
begin
  inherited KeyDown(Key, Shift);

  if TextLoaded then begin
    if Key in [VK_DELETE, VK_BACK, VK_CLEAR, VK_RETURN] then
      FDataLink.Edit;
  end else
    Key := 0;
end;

// ---------------------------------------------------------------------------
procedure TDBColorMemo.KeyPress(var Key: Char);
begin
  inherited KeyPress(Key);

  if TextLoaded then begin
    if (Key in [#32..#255]) and (FDataLink.Field <> nil) and
       (not FDataLink.Field.IsValidChar(Key) or
       (Length(Text) >=  FDataLink.Field.Size)) then
    begin
      MessageBeep(Cardinal(-1));
      Key := #0;
    end;

    case Key of
      ^I, ^J, ^M, #32..#255:
         FDataLink.Edit;
      #27: // <Esc>
         FDataLink.Reset;
    end;
  end else begin
    if not AutoDisplay and (Key = Chr(VK_RETURN)) then
      LoadText;
    Key := #0;
  end;
end;

// ---------------------------------------------------------------------------
procedure TDBColorMemo.WMLButtonDblClk(var Msg: TWMLButtonDblClk);
begin
  if not TextLoaded and not AutoDisplay then
    LoadText
  else //     !
    inherited;
end;

// ---------------------------------------------------------------------------
procedure TDBColorMemo.WMPaste(var Msg: TWMPaste);
begin
  if TextLoaded then begin
    FDataLink.Edit;
    inherited;
  end;
end;

// ---------------------------------------------------------------------------
procedure TDBColorMemo.WMCut(var Msg: TWMCut);
begin
  if TextLoaded then begin
    FDataLink.Edit;
    inherited;
  end;
end;

// ---------------------------------------------------------------------------
procedure TDBColorMemo.CMExit(var Msg: TCMExit);
begin
  try
    if FDataLink.Editing then
      FDataLink.UpdateRecord;
  except
    SelectAll;
    SetFocus;
    raise;
  end;

  inherited;
end;

end. // TDBColorMemo
