// ==========================================================================//
// TFormatter (C)  .. 1998                                           //
//  TFormatter            //
//   !!!                                             //
// ,             //
//        Format  !         //
// ==========================================================================//
unit Format;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Dialogs;

type
  TFormatType = (ftNone, ftOneToOne, ftOneToMany, ftWordWrap);

  TFormatLine = record  //   ()
    OrigLineIndex: Longint;   //     FLines
    BegPos, EndPos: Longint;  // .  .    
    Line: String;             //  
  end;

  PFormatLine = ^TFormatLine;

  TFormatter = class(TList) //   ()
    FLines: TStrings;
    FCanvas: TCanvas;
    FRect: TRect;
    FLeftMargin:  Longint;
    FRightMargin: Longint;
    FFormatType: TFormatType;
    procedure  AddLine(OrigLineIndex: Longint; BegPos, EndPos: Longint);
    function   GetMaxPos(CurX: Longint; LongWord: String): Longint;
    procedure  FormatOneToOne(Sentence: String; Row: Longint);
    procedure  FormatOneToMany(Sentence: String; Row: Longint);
    procedure  FormatWordWrap(Sentence: String; Row: Longint);
  public
    constructor Create(ALines: TStrings; ACanvas: TCanvas; ARect: TRect;
                       ALeftMargin, ARightMargin: Longint);
    destructor  Destroy; override;
    procedure   Assign(Source: TFormatter);
    procedure   DeleteAll;
    function    GetLineInfo(Index: Longint): PFormatLine;
    procedure   Format(FormatType: TFormatType);
    procedure   SetParams(ACanvas: TCanvas; ARect: TRect);
  end;

implementation

{TFormatLines}

constructor TFormatter.Create(ALines: TStrings; ACanvas: TCanvas; ARect: TRect;
                               ALeftMargin, ARightMargin: Longint);
begin
  inherited Create;

  FLines := ALines;
  FCanvas := ACanvas;
  FRect := ARect;
  FLeftMargin  := ALeftMargin;
  FRightMargin := ARightMargin;
  FFormatType := ftNone;
end;

// ---------------------------------------------------------------------------
destructor TFormatter.Destroy;
begin
  DeleteAll;

  inherited Destroy;
end;

// ---------------------------------------------------------------------------
procedure TFormatter.Assign(Source: TFormatter);
var
  I: Longint;
  P: PFormatLine;
begin
  if (Source <> nil) then begin
    DeleteAll;
    FLines       := Source.FLines;
    FCanvas      := Source.FCanvas;
    FRect        := Source.FRect;
    FLeftMargin  := Source.FLeftMargin;
    FRightMargin := Source.FRightMargin;
    FFormatType  := Source.FFormatType;

    for I := 0 to Source.Count-1 do begin
      P := Source.GetLineInfo(I);
      AddLine(P^.OrigLineIndex, P^.BegPos, P^.EndPos);
    end;
  end;
end;

// ---------------------------------------------------------------------------
procedure TFormatter.AddLine(OrigLineIndex: Longint; BegPos, EndPos: Longint);
var
  P: PFormatLine;
begin
  if (OrigLineIndex < 0) or (OrigLineIndex > FLines.Count-1) then
    Exit;

  New(P);
  P^.OrigLineIndex := OrigLineIndex;
  P^.BegPos := BegPos;
  P^.EndPos := EndPos;
  P^.Line := Copy(FLines[OrigLineIndex], BegPos, EndPos - BegPos + 1);
  Add(P);
end;

// ---------------------------------------------------------------------------
procedure TFormatter.DeleteAll;
var
  I: Longint;
begin
  for I := 0 to Count-1 do
    Dispose(Items[I]);
  Clear;
end;

// ---------------------------------------------------------------------------
function  TFormatter.GetLineInfo(Index: Longint): PFormatLine;
begin
  Result := PFormatLine(Items[Index]);
end;


// ---------------------------------------------------------------------------
// .    (      )
//     CurX
// ---------------------------------------------------------------------------
function TFormatter.GetMaxPos(CurX: Longint; LongWord: String): Longint;
var
  I, X, W: Longint;
begin
  Result := 0;
  X := CurX;
  for I := 1 to Length(LongWord) do begin
    W := FCanvas.TextWidth(LongWord[I]);
     if X + W + FRightMargin > FRect.Right then
       break;
     Inc(Result);
     Inc(X, W);
  end;
end;

// ---------------------------------------------------------------------------
procedure TFormatter.FormatOneToOne(Sentence: String; Row: Longint);
var
  Len, BegPos: Longint;
begin
  Len := Length(Sentence);
  BegPos := 1;
  if Len = 0 then BegPos := Len;
  AddLine(Row, BegPos, Len);
end;

// ---------------------------------------------------------------------------
procedure  TFormatter.FormatOneToMany(Sentence: String; Row: Longint);
var
  BegPos, EndPos, Len: Longint;
  LineChunk: String;
begin
  if Sentence = '' then Exit;

  BegPos := 1;
  LineChunk := Sentence;
  Len := Length(Sentence);
  while LineChunk <> '' do begin
    EndPos := BegPos + GetMaxPos(FLeftMargin, LineChunk) - 1;
    AddLine(Row, BegPos, EndPos);
    BegPos := EndPos + 1;
    if BegPos > Len then
      break;
    LineChunk := Copy(Sentence, BegPos, Len);
  end;
end;

// ---------------------------------------------------------------------------
procedure TFormatter.FormatWordWrap(Sentence: String; Row: Longint);
var
 CurWord : String;          //  
 SpacePos,                  //   
 CurX: Longint;             //  X-
 EndOfSentence : Boolean;   //  
 BegOffset,                 // .    
 EndOffset: Longint;        //     
 WordWidth: Longint;
 MaxPos: Longint;
begin
  if Sentence = '' then Exit;

  with FCanvas do begin
    CurX := FLeftMargin;
    BegOffset := 1;
    EndOffset := 0;
    EndOfSentence := False;
    while not EndOfSentence do begin
      //    ,   
      SpacePos := Pos(' ', Sentence);

      if SpacePos > 0 then begin
         //    + 
         CurWord := Copy(Sentence, 1, SpacePos);
         //   
         Sentence := Copy(Sentence, SpacePos + 1, Length(Sentence) - SpacePos);
      end else begin
         //    
         EndOfSentence := True;
         CurWord := Sentence;
         Sentence := '';
      end;

      WordWidth := TextWidth(CurWord);
      //      
      if CurX + WordWidth + FRightMargin > FRect.Right then begin
         if FLeftMargin + WordWidth + FRightMargin > FRect.Right then begin
           MaxPos := GetMaxPos(CurX, CurWord);
           if MaxPos > 0 then begin
             CurWord := Copy(CurWord, MaxPos+1, 255);
             EndOffset := EndOffset + MaxPos;
           end;
         end;

         //     
         AddLine(Row, BegOffset, EndOffset);
         //   .    
         BegOffset := EndOffset + 1;
         CurX := FLeftMargin;
         Sentence := CurWord + Sentence;
         EndOfSentence := False;
      end else begin
        Inc(EndOffset, Length(CurWord));
        Inc(CurX , WordWidth);
      end;
    end;

    //     
    if  BegOffset <= EndOffset then
      AddLine(Row, BegOffset, EndOffset);
  end;
end;

// ---------------------------------------------------------------------------
procedure  TFormatter.Format(FormatType: TFormatType);
var
  I: Longint;
  Sentence: String;
begin
  if FLines.Count = 0 then
    Exit;

  if Count > 0 then
    DeleteAll;

  for I := 0 to FLines.Count-1 do begin
    Sentence := FLines[I];

    if Sentence = '' then
      FormatOneToOne(Sentence, I)
    else
      case FormatType of
        ftOneToOne:  FormatOneToOne(Sentence, I);
        ftOneToMany: FormatOneToMany(Sentence, I);
        ftWordWrap:  FormatWordWrap(Sentence, I);
      end;
  end;
  FFormatType := FormatType;
end;

// ---------------------------------------------------------------------------
procedure TFormatter.SetParams(ACanvas: TCanvas; ARect: TRect);
begin
  FCanvas := ACanvas;
  FRect := ARect;

  if FFormatType <> ftNone then
    Format(FFormatType);
end;

end.
