//===========================================================================//
//                                                                           //
// TSymbolColor -   (C)  .. 1998                      //
//                                                                           //
//===========================================================================//

unit SymColor;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

type
  TSymbolColor = class(TPersistent)
  protected
    FBackColor     : TColor;
    FTextColor     : TColor;
    FUseDefault    : Boolean;

    FOnColorChange : TNotifyEvent;

    cDefBackColor  : TColor;
    cDefTextColor  : TColor;

    procedure SetBackColor(Value: TColor);
    procedure SetTextColor(Value: TColor);
    procedure SetUseDefault(Value: Boolean);

    procedure DoOnColorChange; dynamic;
    procedure ResetToDefaultColors;
  public
    constructor Create(FG, BG : TColor); virtual;
    procedure Assign(Source : TPersistent); override;
    property OnColorChange : TNotifyEvent read FOnColorChange write FOnColorChange;
  published
    property BackColor : TColor   read FBackColor  write SetBackColor;
    property TextColor : TColor   read FTextColor  write SetTextColor;
    property UseDefault : Boolean read FUseDefault write SetUseDefault;
  end; // TSymbolColor

implementation

{TSymbolColor}

// ---------------------------------------------------------------------------
constructor TSymbolColor.Create(FG, BG : TColor);
begin
  inherited Create;

  cDefBackColor := BG;
  cDefTextColor := FG;
  FUseDefault   := True;

  ResetToDefaultColors;
end;

// ---------------------------------------------------------------------------
procedure TSymbolColor.DoOnColorChange;
begin
  if Assigned(FOnColorChange) then
    FOnColorChange(Self);
end;

// ---------------------------------------------------------------------------
procedure TSymbolColor.ResetToDefaultColors;
begin
  FBackColor := cDefBackColor;
  FTextColor := cDefTextColor;
end;

// ---------------------------------------------------------------------------
procedure TSymbolColor.SetBackColor(Value: TColor);
begin
  if Value <> FBackColor then begin
    if Value <> cDefBackColor then
      FUseDefault := False;
    FBackColor := Value;
    DoOnColorChange;
  end;
end;

// ---------------------------------------------------------------------------
procedure TSymbolColor.SetTextColor(Value: TColor);
begin
  if Value <> FTextColor then begin
    if Value <> cDefTextColor then
      FUseDefault := False;
    FTextColor := Value;
    DoOnColorChange;
  end;
end;

// ---------------------------------------------------------------------------
procedure TSymbolColor.SetUseDefault(Value: Boolean);
begin
  FUseDefault := Value;
  if FUseDefault then begin
    ResetToDefaultColors;
    DoOnColorChange;
  end;
end;

// ---------------------------------------------------------------------------
procedure TSymbolColor.Assign(Source : TPersistent);
var
  C: TSymbolColor absolute Source;
begin
  if (Source <> nil) and (Source is TSymbolColor) then begin
    BackColor  := C.BackColor;
    TextColor  := C.TextColor;
  end else
    inherited Assign(Source);
end;

end.

